# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test_read_efficiency_with_streamer():
    with uproot.open(skhep_testdata.data_path("uproot-issue38c.root")) as fp:
        eff = fp["TEfficiencyName"]
        assert eff
        assert [pair.members for pair in eff.members["fBeta_bin_params"]] == [
            {"first": -1.0, "second": -2.0},
            {"first": 2.0, "second": 4.0},
            {"first": 4.0, "second": 8.0},
            {"first": 8.0, "second": 16.0},
            {"first": 16.0, "second": 32.0},
            {"first": 32.0, "second": 64.0},
            {"first": 64.0, "second": 128.0},
            {"first": 128.0, "second": 256.0},
            {"first": 256.0, "second": 512.0},
            {"first": 512.0, "second": 1024.0},
            {"first": 1024.0, "second": 2048.0},
            {"first": -1.0, "second": -2.0},
            {"first": 1.0, "second": 1.0},
        ]


def test_read_efficiency_without_streamer(reset_classes):
    with uproot.open(skhep_testdata.data_path("uproot-issue209.root")) as fp:
        with pytest.raises(NotImplementedError):
            assert fp["TEfficiencyName"]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pickle
import sys

import numpy as np
import pytest
import skhep_testdata


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="Dynamic types depend on module __getattr__, a Python 3.7+ feature.",
)
def test():
    with open("tests/samples/h_dynamic.pkl", "rb") as f:
        assert len(list(pickle.load(f).axis(0))) == 100



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines versioned models for ``TDatime``.
"""


import struct

import numpy

import uproot
import uproot._awkward_forth
import uproot.behaviors.TDatime

_tdatime_format1 = struct.Struct(">I")


class Model_TDatime(uproot.behaviors.TDatime.TDatime, uproot.model.Model):
    """
    A versionless :doc:`uproot.model.Model` for ``TDatime``.
    """

    def read_numbytes_version(self, chunk, cursor, context):
        pass

    def read_members(self, chunk, cursor, context, file):
        forth_stash = uproot._awkward_forth.forth_stash(context)
        if forth_stash is not None:
            forth_obj = forth_stash.get_gen_obj()
            key = forth_obj.get_keys(1)
            form_key = f"node{key}-data"
            forth_stash.add_to_header(f"output node{key}-data int32\n")
            forth_stash.add_to_pre(f"stream !I-> node{key}-data\n")
            form_key = f"node{key}-data"
            if forth_obj.should_add_form():
                forth_obj.add_form_key(form_key)
                temp_aform = {
                    "class": "RecordArray",
                    "contents": {
                        "fDatime": {
                            "class": "NumpyArray",
                            "primitive": "uint32",
                            "form_key": f"node{key}",
                        }
                    },
                    "parameters": {"__record__": "TDatime"},
                }
                forth_obj.add_form(temp_aform)
            temp_form = forth_obj.add_node(
                f"node{key}",
                forth_stash.get_attrs(),
                "i64",
                0,
                None,
            )
            forth_obj.go_to(temp_form)
        self._members["fDatime"] = cursor.field(chunk, _tdatime_format1, context)

    @classmethod
    def strided_interpretation(
        cls, file, header=False, tobject_header=True, breadcrumbs=(), original=None
    ):
        members = []
        if header:
            members.append(("@num_bytes", numpy.dtype(">u4")))
            members.append(("@instance_version", numpy.dtype(">u2")))
        members.append(("fDatime", numpy.dtype(">u4")))
        return uproot.interpretation.objects.AsStridedObjects(
            cls, members, original=original
        )

    @classmethod
    def awkward_form(cls, file, context):
        awkward = uproot.extras.awkward()
        contents = {}
        if context["header"]:
            contents["@num_bytes"] = uproot._util.awkward_form(
                numpy.dtype("u4"), file, context
            )
            contents["@instance_version"] = uproot._util.awkward_form(
                numpy.dtype("u2"), file, context
            )
        contents["fDatime"] = uproot._util.awkward_form(
            numpy.dtype(">u4"), file, context
        )
        return awkward.forms.RecordForm(
            list(contents.values()),
            list(contents.keys()),
            parameters={"__record__": "TDatime"},
        )

    base_names_versions = []
    member_names = ["fDatime"]
    class_flags = {}
    class_code = None


uproot.classes["TDatime"] = Model_TDatime



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines languages for expressions passed to
:ref:`uproot.behaviors.TBranch.HasBranches.arrays` (and similar).

The default is :doc:`uproot.language.python.PythonLanguage`.

All languages must be subclasses of :doc:`uproot.language.Language`.
"""


class Language:
    """
    Abstract class for all languages, which are used to compute the expressions
    that are passed to :ref:`uproot.behaviors.TBranch.HasBranches.arrays` (and
    similar).

    The default is :doc:`uproot.language.python.PythonLanguage`.
    """

    pass



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import array
import os

import numpy as np
import pytest
import skhep_testdata

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test_2dim(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        tree = fout.mktree("tree", {"branch": np.dtype((np.float64, (3,)))})
        tree.extend({"branch": np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])})
        with pytest.raises(ValueError):
            tree.extend({"branch": np.array([7.7, 8.8, 9.9])})
        with pytest.raises(ValueError):
            tree.extend({"branch": np.array([[7.7], [8.8], [9.9]])})

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert [np.asarray(x.branch).tolist() for x in t1] == [
        [1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6],
    ]

    with uproot.open(newfile) as fin:
        assert fin["tree/branch"].array().tolist() == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]

    f1.Close()


def test_2dim_interface(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = {"branch": np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])}
        fout["tree"].extend(
            {"branch": np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])}
        )
        with pytest.raises(ValueError):
            fout["tree"].extend({"branch": np.array([7.7, 8.8, 9.9])})
        with pytest.raises(ValueError):
            fout["tree"].extend({"branch": np.array([[7.7], [8.8], [9.9]])})

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert [np.asarray(x.branch).tolist() for x in t1] == [
        [1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6],
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [7.0, 8.0, 9.0],
    ]

    with uproot.open(newfile) as fin:
        assert fin["tree/branch"].array().tolist() == [
            [1.1, 2.2, 3.3],
            [4.4, 5.5, 6.6],
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
        ]

    f1.Close()


def test_2dim_interface_2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = {"branch": [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]}
        fout["tree"].extend(
            {"branch": [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]}
        )
        with pytest.raises(ValueError):
            fout["tree"].extend({"branch": [7.7, 8.8, 9.9]})
        with pytest.raises(ValueError):
            fout["tree"].extend({"branch": [[7.7], [8.8], [9.9]]})

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert [np.asarray(x.branch).tolist() for x in t1] == [
        [1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6],
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [7.0, 8.0, 9.0],
    ]

    with uproot.open(newfile) as fin:
        assert fin["tree/branch"].array().tolist() == [
            [1.1, 2.2, 3.3],
            [4.4, 5.5, 6.6],
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
        ]

    f1.Close()


def test_structured_array(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = np.array(
            [(1, 1.1), (2, 2.2), (3, 3.3)], [("x", np.int32), ("y", np.float64)]
        )
        fout["tree"].extend(
            np.array(
                [(4, 4.4), (5, 5.5), (6, 6.6)], [("x", np.int32), ("y", np.float64)]
            )
        )

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert t1.GetBranch("x").GetName() == "x"
    assert t1.GetBranch("y").GetName() == "y"
    assert [np.asarray(x.x).tolist() for x in t1] == [1, 2, 3, 4, 5, 6]
    assert [np.asarray(x.y).tolist() for x in t1] == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    with uproot.open(newfile) as fin:
        assert fin["tree/x"].name == "x"
        assert fin["tree/y"].name == "y"
        assert fin["tree/x"].typename == "int32_t"
        assert fin["tree/y"].typename == "double"
        assert fin["tree/x"].array().tolist() == [1, 2, 3, 4, 5, 6]
        assert fin["tree/y"].array().tolist() == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    f1.Close()


def test_pandas(tmp_path):
    pandas = pytest.importorskip("pandas")

    newfile = os.path.join(tmp_path, "newfile.root")

    df1 = pandas.DataFrame({"x": [1, 2, 3], "y": [1.1, 2.2, 3.3]})
    df2 = pandas.DataFrame({"x": [4, 5, 6], "y": [4.4, 5.5, 6.6]})

    with uproot.recreate(newfile, compression=None) as fout:
        fout["tree"] = df1
        fout["tree"].extend(df2)

    f1 = ROOT.TFile(newfile)
    t1 = f1.Get("tree")

    assert t1.GetBranch("index").GetName() == "index"
    assert t1.GetBranch("x").GetName() == "x"
    assert t1.GetBranch("y").GetName() == "y"
    assert [np.asarray(x.x).tolist() for x in t1] == [1, 2, 3, 4, 5, 6]
    assert [np.asarray(x.y).tolist() for x in t1] == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    with uproot.open(newfile) as fin:
        assert fin["tree/index"].name == "index"
        assert fin["tree/x"].name == "x"
        assert fin["tree/y"].name == "y"
        assert fin["tree/index"].typename.startswith("int")
        assert fin["tree/x"].typename.startswith("int")
        assert fin["tree/y"].typename == "double"
        assert fin["tree/x"].array().tolist() == [1, 2, 3, 4, 5, 6]
        assert fin["tree/y"].array().tolist() == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

    f1.Close()


def test_histogram_interface(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as fin:
        h1d, h2d = fin["hpx"], fin["hpxpy"]
        (h1d_entries_1, h1d_xedges_1) = h1d.to_numpy()
        (h2d_entries_1, h2d_xedges_1, h2d_yedges_1) = h2d.to_numpy(dd=False)
        (h2d_dd_entries_1, (h2d_dd_xedges_1, h2d_dd_yedges_1)) = h2d.to_numpy(dd=True)

        with uproot.recreate(newfile) as fout:
            fout["h1d"] = h1d.to_numpy()
            fout["h2d"] = h2d.to_numpy(dd=False)
            fout["h2d_dd"] = h2d.to_numpy(dd=True)
            (h1d_entries_2, h1d_xedges_2) = fout["h1d"].to_numpy()
            (h2d_entries_2, h2d_xedges_2, h2d_yedges_2) = fout["h2d"].to_numpy(dd=False)
            (h2d_dd_entries_2, (h2d_dd_xedges_2, h2d_dd_yedges_2)) = fout[
                "h2d_dd"
            ].to_numpy(dd=True)

    with uproot.open(newfile) as finagin:
        assert np.array_equal(h1d_entries_1, h1d_entries_2)
        assert np.array_equal(h1d_xedges_1, h1d_xedges_2)

        assert np.array_equal(h2d_entries_1, h2d_entries_2)
        assert np.array_equal(h2d_xedges_1, h2d_xedges_2)
        assert np.array_equal(h2d_yedges_1, h2d_yedges_2)

        assert np.array_equal(h2d_dd_entries_1, h2d_dd_entries_2)
        assert np.array_equal(h2d_dd_xedges_1, h2d_dd_xedges_2)
        assert np.array_equal(h2d_dd_yedges_1, h2d_dd_yedges_2)


def test_ex_nihilo_TH1(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    h1 = uproot.writing.identify.to_TH1x(
        fName="h1",
        fTitle="title",
        data=np.array([0, 2, 5, 0], np.float64),
        fEntries=7,
        fTsumw=1,
        fTsumw2=1,
        fTsumwx=1,
        fTsumwx2=1,
        fSumw2=np.array([0, 2, 5, 0], np.float64),
        fXaxis=uproot.writing.identify.to_TAxis(
            fName="xaxis",
            fTitle="",
            fNbins=2,
            fXmin=-3.14,
            fXmax=2.71,
        ),
    )

    with uproot.recreate(newfile) as fout:
        fout["out"] = h1
        fout["there"] = h1.to_numpy()

    f3 = ROOT.TFile(newfile)
    for name in "out", "there":
        h3 = f3.Get(name)
        assert h3.GetEntries() == pytest.approx(7)
        assert h3.GetBinLowEdge(1) == pytest.approx(-3.14)
        assert h3.GetBinWidth(1) == pytest.approx((2.71 - -3.14) / 2)
        assert h3.GetBinContent(0) == pytest.approx(0)
        assert h3.GetBinContent(1) == pytest.approx(2)
        assert h3.GetBinContent(2) == pytest.approx(5)
        assert h3.GetBinContent(3) == pytest.approx(0)
        assert h3.GetBinError(0) == pytest.approx(0)
        assert h3.GetBinError(1) == pytest.approx(np.sqrt(2))
        assert h3.GetBinError(2) == pytest.approx(np.sqrt(5))
        assert h3.GetBinError(3) == pytest.approx(0)

    f3.Close()


def test_ex_nihilo_TH2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    h1 = uproot.writing.identify.to_TH2x(
        fName="h1",
        fTitle="title",
        data=np.array(
            [0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0], np.float64
        ),
        fEntries=7,
        fTsumw=1,
        fTsumw2=1,
        fTsumwx=1,
        fTsumwx2=1,
        fTsumwy=-1,
        fTsumwy2=1,
        fTsumwxy=-1,
        fSumw2=np.array(
            [0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0], np.float64
        ),
        fXaxis=uproot.writing.identify.to_TAxis(
            fName="xaxis",
            fTitle="",
            fNbins=2,
            fXmin=-3.14,
            fXmax=2.71,
        ),
        fYaxis=uproot.writing.identify.to_TAxis(
            fName="yaxis",
            fTitle="",
            fNbins=3,
            fXmin=-5.0,
            fXmax=10.0,
        ),
    )

    with uproot.recreate(newfile) as fout:
        fout["out"] = h1
        fout["there"] = h1.to_numpy()

    f3 = ROOT.TFile(newfile)
    for name in "out", "there":
        h3 = f3.Get(name)
        assert h3.GetEntries() == 7
        assert h3.GetNbinsX() == 2
        assert h3.GetNbinsY() == 3
        assert h3.GetXaxis().GetBinLowEdge(1) == pytest.approx(-3.14)
        assert h3.GetXaxis().GetBinUpEdge(2) == pytest.approx(2.71)
        assert h3.GetYaxis().GetBinLowEdge(1) == pytest.approx(-5)
        assert h3.GetYaxis().GetBinUpEdge(3) == pytest.approx(10)
        assert [[h3.GetBinContent(i, j) for j in range(5)] for i in range(4)] == [
            pytest.approx([0, 0, 0, 0, 0]),
            pytest.approx([0, 0, 0, 2, 0]),
            pytest.approx([0, 5, 0, 0, 0]),
            pytest.approx([0, 0, 0, 0, 0]),
        ]

    f3.Close()


def test_ex_nihilo_TH3(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    h1 = uproot.writing.identify.to_TH3x(
        fName="h1",
        fTitle="title",
        data=np.array(
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            + [0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0]
            + [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            np.float64,
        ),
        fEntries=7,
        fTsumw=1,
        fTsumw2=1,
        fTsumwx=1,
        fTsumwx2=1,
        fTsumwy=1,
        fTsumwy2=1,
        fTsumwxy=1,
        fTsumwz=1,
        fTsumwz2=1,
        fTsumwxz=1,
        fTsumwyz=1,
        fSumw2=np.array(
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            + [0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0]
            + [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            np.float64,
        ),
        fXaxis=uproot.writing.identify.to_TAxis(
            fName="xaxis",
            fTitle="",
            fNbins=2,
            fXmin=-3.14,
            fXmax=2.71,
        ),
        fYaxis=uproot.writing.identify.to_TAxis(
            fName="yaxis",
            fTitle="",
            fNbins=3,
            fXmin=-5.0,
            fXmax=10.0,
        ),
        fZaxis=uproot.writing.identify.to_TAxis(
            fName="zaxis",
            fTitle="",
            fNbins=1,
            fXmin=100.0,
            fXmax=200.0,
        ),
    )

    with uproot.recreate(newfile) as fout:
        fout["out"] = h1
        fout["there"] = h1.to_numpy()

    f3 = ROOT.TFile(newfile)
    for name in "out", "there":
        h3 = f3.Get(name)
        assert h3.GetEntries() == 7
        assert h3.GetNbinsX() == 2
        assert h3.GetNbinsY() == 3
        assert h3.GetNbinsZ() == 1
        assert h3.GetXaxis().GetBinLowEdge(1) == pytest.approx(-3.14)
        assert h3.GetXaxis().GetBinUpEdge(2) == pytest.approx(2.71)
        assert h3.GetYaxis().GetBinLowEdge(1) == pytest.approx(-5)
        assert h3.GetYaxis().GetBinUpEdge(3) == pytest.approx(10)
        assert h3.GetZaxis().GetBinLowEdge(1) == pytest.approx(100)
        assert h3.GetZaxis().GetBinUpEdge(1) == pytest.approx(200)
        approx = pytest.approx
        assert [
            [[h3.GetBinContent(i, j, k) for k in range(3)] for j in range(5)]
            for i in range(4)
        ] == [
            [[0, 0, 0], approx([0, 0, 0]), [0, 0, 0], approx([0, 0, 0]), [0, 0, 0]],
            [[0, 0, 0], approx([0, 0, 0]), [0, 0, 0], approx([0, 2, 0]), [0, 0, 0]],
            [[0, 0, 0], approx([0, 5, 0]), [0, 0, 0], approx([0, 0, 0]), [0, 0, 0]],
            [[0, 0, 0], approx([0, 0, 0]), [0, 0, 0], approx([0, 0, 0]), [0, 0, 0]],
        ]

    f3.Close()


def test_fromroot_TH1(tmp_path):
    original = os.path.join(tmp_path, "original.root")
    newfile = os.path.join(tmp_path, "newfile.root")

    f1 = ROOT.TFile(original, "recreate")
    h1 = ROOT.TH1D("h1", "title", 2, 3.0, 5.0)  # centers: 3.5 4.5
    h1.Sumw2()
    h1.Fill(3.5)
    h1.Fill(3.5)
    h1.Fill(4.5)
    h1.Write()
    f1.Close()

    with uproot.open(original) as fin:
        h2 = fin["h1"]

        with uproot.recreate(newfile) as fout:
            fout["out"] = h2
            fout["there"] = h2.to_numpy()

    with uproot.open(newfile) as finagin:
        assert np.array_equal(finagin["out"].member("fSumw2"), [0, 2, 1, 0])
        assert np.array_equal(finagin["there"].member("fSumw2"), [])
        assert {
            k: v for k, v in finagin["out"].all_members.items() if k.startswith("fTs")
        } == {
            k: v for k, v in finagin["there"].all_members.items() if k.startswith("fTs")
        }

    f1.Close()


def test_fromroot_TH2(tmp_path):
    original = os.path.join(tmp_path, "original.root")
    newfile = os.path.join(tmp_path, "newfile.root")

    f1 = ROOT.TFile(original, "recreate")
    h1 = ROOT.TH2D(
        "h1", "title", 2, 3.0, 5.0, 3, 10, 13
    )  # centers: 3.5 4.5; 10.5 11.5 12.5
    h1.Sumw2()
    h1.Fill(3.5, 10.5)
    h1.Fill(3.5, 10.5)
    h1.Fill(4.5, 11.5)
    h1.Fill(4.5, 12.5)

    h1.Write()
    f1.Close()

    with uproot.open(original) as fin:
        h2 = fin["h1"]

        with uproot.recreate(newfile) as fout:
            fout["out"] = h2
            fout["there"] = h2.to_numpy()

    with uproot.open(newfile) as finagin:
        assert np.array_equal(
            finagin["out"].member("fSumw2"),
            [0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0]
            + [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        )
        assert np.array_equal(finagin["there"].member("fSumw2"), [])
        assert {
            k: v for k, v in finagin["out"].all_members.items() if k.startswith("fTs")
        } == {
            k: v for k, v in finagin["there"].all_members.items() if k.startswith("fTs")
        }

    f1.Close()


def test_fromroot_TH3(tmp_path):
    original = os.path.join(tmp_path, "original.root")
    newfile = os.path.join(tmp_path, "newfile.root")

    f1 = ROOT.TFile(original, "recreate")
    h1 = ROOT.TH3D(
        "h1", "title", 2, 3.0, 5.0, 3, 10, 13, 4, 100, 104
    )  # centers: 3.5 4.5; 10.5 11.5 12.5; 100.5, 101.5, 102.5, 103.5
    h1.Sumw2()
    h1.Fill(3.5, 10.5, 100.5)
    h1.Fill(3.5, 10.5, 100.5)
    h1.Fill(3.5, 11.5, 102.5)
    h1.Fill(3.5, 12.5, 101.5)
    h1.Write()
    f1.Close()

    with uproot.open(original) as fin:
        h2 = fin["h1"]

        with uproot.recreate(newfile) as fout:
            fout["out"] = h2
            fout["there"] = h2.to_numpy()

    with uproot.open(newfile) as finagin:
        assert np.array_equal(
            finagin["out"].member("fSumw2"),
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        )
        assert np.array_equal(finagin["there"].member("fSumw2"), [])
        assert {
            k: v for k, v in finagin["out"].all_members.items() if k.startswith("fTs")
        } == {
            k: v for k, v in finagin["there"].all_members.items() if k.startswith("fTs")
        }

    f1.Close()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test():
    with uproot.open(skhep_testdata.data_path("uproot-issue-750.root"))[
        "tout/Electron_eta"
    ] as branch:
        branch.array(library="np")



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import skhep_testdata
from numpy.testing import assert_array_equal

import uproot


def test():
    with uproot.open(skhep_testdata.data_path("uproot-issue64.root")) as file:
        p = file["events/nbevents"]
        assert p.value == 500
        assert p
        assert int(p) == 500
        assert float(p) == 500.0


def test_issue_707():
    with uproot.open(skhep_testdata.data_path("uproot-issue-707.root")) as file:
        p = file["NumberOfPrimariesEdep"]
        assert p.value == 100000004



from collections.abc import Callable, Iterable, Mapping

import numpy

import uproot
from uproot._util import no_filter
from uproot.behaviors.TBranch import HasBranches, TBranch, _regularize_step_size


def dask(
    files,
    *,
    filter_name=no_filter,
    filter_typename=no_filter,
    filter_branch=no_filter,
    recursive=True,
    full_paths=False,
    step_size="100 MB",
    library="ak",
    ak_add_doc=False,
    custom_classes=None,
    allow_missing=False,
    open_files=True,
    **options,
):
    """
    Args:
        files: See below.
        filter_name (None, glob string, regex string in ``"/pattern/i"`` syntax, function of str \u2192 bool, or iterable of the above): A
            filter to select ``TBranches`` by name.
        filter_typename (None, glob string, regex string in ``"/pattern/i"`` syntax, function of str \u2192 bool, or iterable of the above): A
            filter to select ``TBranches`` by type.
        filter_branch (None or function of :doc:`uproot.behaviors.TBranch.TBranch` \u2192 bool, :doc:`uproot.interpretation.Interpretation`, or None): A
            filter to select ``TBranches`` using the full
            :doc:`uproot.behaviors.TBranch.TBranch` object. If the function
            returns False or None, the ``TBranch`` is excluded; if the function
            returns True, it is included with its standard
            :ref:`uproot.behaviors.TBranch.TBranch.interpretation`; if an
            :doc:`uproot.interpretation.Interpretation`, this interpretation
            overrules the standard one.
        recursive (bool): If True, include all subbranches of branches as
            separate fields; otherwise, only search one level deep.
        full_paths (bool): If True, include the full path to each subbranch
            with slashes (``/``); otherwise, use the descendant's name as
            the field name.
        step_size (int or str): If an integer, the maximum number of entries to
            include in each chunk; if a string, the maximum memory_size to include
            in each chunk. The string must be a number followed by a memory unit,
            such as "100 MB".
        library (str or :doc:`uproot.interpretation.library.Library`): The library
            that is used to represent arrays. If ``library='np'`` it returns a dict
            of dask arrays and if ``library='ak'`` it returns a single dask-awkward
            array. ``library='pd'`` has not been implemented yet and will raise a
            ``NotImplementedError``.
        ak_add_doc (bool): If True and ``library="ak"``, add the TBranch ``title``
            to the Awkward ``__doc__`` parameter of the array.
        custom_classes (None or dict): If a dict, override the classes from
            the :doc:`uproot.reading.ReadOnlyFile` or ``uproot.classes``.
        allow_missing (bool): If True, skip over any files that do not contain
            the specified ``TTree``.
        open_files (bool): If True (default), the function will open the files to read file
            metadata, i.e. only the main data read is delayed till the compute call on
            the dask collections. If False, the opening of the files and reading the
            metadata is also delayed till the compute call. In this case, branch-names
            are inferred by opening only the first file.
        options: See below.

    Returns dask equivalents of the backends supported by uproot. If ``library='np'``,
    the function returns a Python dict of dask arrays. If ``library='ak'``, the
    function returns a single dask-awkward array.

    For example:

    .. code-block:: python

        >>> uproot.dask(root_file)
        dask.awkward<from-uproot, npartitions=1>
        >>> uproot.dask(root_file,library='np')
        {'Type': dask.array<Type-from-uproot, shape=(2304,), dtype=object, chunksize=(2304,), chunktype=numpy.ndarray>, ...}


    This function (naturally) depends on Dask. To use it with ``library="np"``:

    .. code-block:: bash

        # with pip
        pip install "dask[complete]"
        # or with conda
        conda install dask

    For using ``library='ak'``

    .. code-block:: bash

        pip install dask-awkward   # not on conda-forge yet

    Allowed types for the ``files`` parameter:

    * str/bytes: relative or absolute filesystem path or URL, without any colons
      other than Windows drive letter or URL schema.
      Examples: ``"rel/file.root"``, ``"C:\\abs\\file.root"``, ``"http://where/what.root"``
    * str/bytes: same with an object-within-ROOT path, separated by a colon.
      Example: ``"rel/file.root:tdirectory/ttree"``
    * pathlib.Path: always interpreted as a filesystem path or URL only (no
      object-within-ROOT path), regardless of whether there are any colons.
      Examples: ``Path("rel:/file.root")``, ``Path("/abs/path:stuff.root")``
    * glob syntax in str/bytes and pathlib.Path.
      Examples: ``Path("rel/*.root")``, ``"/abs/*.root:tdirectory/ttree"``
    * dict: keys are filesystem paths, values are objects-within-ROOT paths.
      Example: ``{{"/data_v1/*.root": "ttree_v1", "/data_v2/*.root": "ttree_v2"}}``
    * already-open TTree objects.
    * iterables of the above.

    Options (type; default):

    * file_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.file.MemmapSource`)
    * xrootd_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.xrootd.XRootDSource`)
    * http_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.http.HTTPSource`)
    * object_handler (:doc:`uproot.source.chunk.Source` class; :doc:`uproot.source.object.ObjectSource`)
    * timeout (float for HTTP, int for XRootD; 30)
    * max_num_elements (None or int; None)
    * num_workers (int; 1)
    * num_fallback_workers (int; 10)
    * begin_chunk_size (memory_size; 512)
    * minimal_ttree_metadata (bool; True)

    Other file entry points:

    * :doc:`uproot.reading.open`: opens one file to read any of its objects.
    * :doc:`uproot.behaviors.TBranch.iterate`: iterates through chunks of
      contiguous entries in ``TTrees``.
    * :doc:`uproot.behaviors.TBranch.concatenate`: returns a single
      concatenated array from ``TTrees``.
    * :doc:`uproot._dask.dask` (this function): returns an unevaluated Dask
      array from ``TTrees``.
    """

    files = uproot._util.regularize_files(files)
    library = uproot.interpretation.library._regularize_library(library)

    if library.name == "pd":
        raise NotImplementedError()

    real_options = options.copy()
    if "num_workers" not in real_options:
        real_options["num_workers"] = 1
    if "num_fallback_workers" not in real_options:
        real_options["num_fallback_workers"] = 1

    filter_branch = uproot._util.regularize_filter(filter_branch)

    interp_options = {"ak_add_doc": ak_add_doc}

    if library.name == "np":
        if open_files:
            return _get_dask_array(
                files,
                filter_name,
                filter_typename,
                filter_branch,
                recursive,
                full_paths,
                step_size,
                custom_classes,
                allow_missing,
                real_options,
                interp_options,
            )
        else:
            return _get_dask_array_delay_open(
                files,
                filter_name,
                filter_typename,
                filter_branch,
                recursive,
                full_paths,
                custom_classes,
                allow_missing,
                real_options,
                interp_options,
            )
    elif library.name == "ak":
        if open_files:
            return _get_dak_array(
                files,
                filter_name,
                filter_typename,
                filter_branch,
                recursive,
                full_paths,
                step_size,
                custom_classes,
                allow_missing,
                real_options,
                interp_options,
            )
        else:
            return _get_dak_array_delay_open(
                files,
                filter_name,
                filter_typename,
                filter_branch,
                recursive,
                full_paths,
                custom_classes,
                allow_missing,
                real_options,
                interp_options,
            )
    else:
        raise NotImplementedError()


class _PackedArgCallable:
    """Wrap a callable such that packed arguments can be unrolled.
    Inspired by dask.dataframe.io.io._PackedArgCallable.
    """

    def __init__(
        self,
        func: Callable,
        args=None,
        kwargs=None,
        packed: bool = False,
    ):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.packed = packed

    def __call__(self, packed_arg):
        if not self.packed:
            packed_arg = (packed_arg,)
        return self.func(
            *packed_arg,
            *(self.args or []),
            **(self.kwargs or {}),
        )


class _LazyInputsDict(Mapping):
    """Dictionary with lazy key value pairs
    Parameters
    ----------
    inputs : list[Any]
        The list of dicionary values.
    """

    def __init__(self, inputs, **kwargs) -> None:
        self.inputs = inputs
        self.kwargs = kwargs

    def __len__(self):
        return len(self.inputs)

    def __iter__(self):
        return (self[k] for k in self.keys())

    def __getitem__(self, i):
        return self.inputs[i[0]]

    def __contains__(self, k):
        if isinstance(k, tuple):
            if isinstance(k[0], int):
                return k[0] >= 0 and k[0] < len(self)
        return False

    def keys(self):
        return ((i,) for i in range(len(self.inputs)))


def _dask_array_from_map(
    func,
    *iterables,
    chunks,
    dtype,
    args=None,
    label=None,
    token=None,
    **kwargs,
):
    dask = uproot.extras.dask()
    da = uproot.extras.dask_array()
    if not callable(func):
        raise ValueError("`func` argument must be `callable`")
    lengths = set()
    iters = list(iterables)
    for i, iterable in enumerate(iters):
        if not isinstance(iterable, Iterable):
            raise ValueError(
                f"All elements of `iterables` must be Iterable, got {type(iterable)}"
            )
        try:
            lengths.add(len(iterable))  # type: ignore
        except (AttributeError, TypeError):
            iters[i] = list(iterable)
            lengths.add(len(iters[i]))  # type: ignore
    if len(lengths) == 0:
        raise ValueError("`from_map` requires at least one Iterable input")
    elif len(lengths) > 1:
        raise ValueError("All `iterables` must have the same length")
    if lengths == {0}:
        raise ValueError("All `iterables` must have a non-zero length")

    # Check for `produces_tasks` and `creation_info`
    produces_tasks = kwargs.pop("produces_tasks", False)
    # creation_info = kwargs.pop("creation_info", None)

    if produces_tasks or len(iters) == 1:
        if len(iters) > 1:
            # Tasks are not detected correctly when they are "packed"
            # within an outer list/tuple
            raise ValueError(
                "Multiple iterables not supported when produces_tasks=True"
            )
        inputs = list(iters[0])
        packed = False
    else:
        # Structure inputs such that the tuple of arguments pair each 0th,
        # 1st, 2nd, ... elements together; for example:
        # from_map(f, [1, 2, 3], [4, 5, 6]) --> [f(1, 4), f(2, 5), f(3, 6)]
        inputs = list(zip(*iters))
        packed = True

    # Define collection name
    label = label or dask.utils.funcname(func)
    token = token or dask.base.tokenize(func, iters, **kwargs)
    name = f"{label}-{token}"

    # Define io_func
    if packed or args or kwargs:
        io_func = _PackedArgCallable(
            func,
            args=args,
            kwargs=kwargs,
            packed=packed,
        )
    else:
        io_func = func

    io_arg_map = dask.blockwise.BlockwiseDepDict(
        mapping=_LazyInputsDict(inputs),  # type: ignore
        produces_tasks=produces_tasks,
    )

    dsk = dask.blockwise.Blockwise(
        output=name,
        output_indices="i",
        dsk={name: (io_func, dask.blockwise.blockwise_token(0))},
        indices=[(io_arg_map, "i")],
        numblocks={},
        annotations=None,
    )

    hlg = dask.highlevelgraph.HighLevelGraph.from_collections(name, dsk)
    return da.core.Array(hlg, name, chunks, dtype=dtype)


class _UprootReadNumpy:
    def __init__(self, ttrees, key, interp_options) -> None:
        self.ttrees = ttrees
        self.key = key
        self.interp_options = interp_options

    def __call__(self, i_start_stop):
        i, start, stop = i_start_stop
        return self.ttrees[i][self.key].array(
            entry_start=start,
            entry_stop=stop,
            library="np",
            ak_add_doc=self.interp_options["ak_add_doc"],
        )


class _UprootOpenAndReadNumpy:
    def __init__(
        self, custom_classes, allow_missing, real_options, key, interp_options
    ):
        self.custom_classes = custom_classes
        self.allow_missing = allow_missing
        self.real_options = real_options
        self.key = key
        self.interp_options = interp_options

    def __call__(self, file_path_object_path):
        file_path, object_path = file_path_object_path
        ttree = uproot._util.regularize_object_path(
            file_path,
            object_path,
            self.custom_classes,
            self.allow_missing,
            self.real_options,
        )
        return ttree[self.key].array(
            library="np", ak_add_doc=self.interp_options["ak_add_doc"]
        )


def _get_dask_array(
    files,
    filter_name=no_filter,
    filter_typename=no_filter,
    filter_branch=no_filter,
    recursive=True,
    full_paths=False,
    step_size="100 MB",
    custom_classes=None,
    allow_missing=False,
    real_options=None,
    interp_options=None,
):
    ttrees = []
    common_keys = None
    is_self = []

    count = 0
    for file_path, object_path in files:
        obj = uproot._util.regularize_object_path(
            file_path, object_path, custom_classes, allow_missing, real_options
        )

        if obj is not None:
            count += 1

            if isinstance(obj, TBranch) and len(obj.keys(recursive=True)) == 0:
                original = obj
                obj = obj.parent
                is_self.append(True)

                def real_filter_branch(branch):
                    return branch is original and filter_branch(branch)  # noqa: B023

            else:
                is_self.append(False)
                real_filter_branch = filter_branch

            ttrees.append(obj)

            new_keys = obj.keys(
                recursive=recursive,
                filter_name=filter_name,
                filter_typename=filter_typename,
                filter_branch=real_filter_branch,
                full_paths=full_paths,
            )

            if common_keys is None:
                common_keys = new_keys
            else:
                new_keys = set(new_keys)
                common_keys = [key for key in common_keys if key in new_keys]

    if count == 0:
        raise ValueError(
            "allow_missing=True and no TTrees found in\n\n    {}".format(
                "\n    ".join(
                    "{"
                    + "{}: {}".format(
                        repr(f.file_path if isinstance(f, HasBranches) else f),
                        repr(f.object_path if isinstance(f, HasBranches) else o),
                    )
                    + "}"
                    for f, o in files
                )
            )
        )

    if len(common_keys) == 0 or not (all(is_self) or not any(is_self)):
        raise ValueError(
            "TTrees in\n\n    {}\n\nhave no TBranches in common".format(
                "\n    ".join(
                    "{"
                    + "{}: {}".format(
                        repr(f.file_path if isinstance(f, HasBranches) else f),
                        repr(f.object_path if isinstance(f, HasBranches) else o),
                    )
                    + "}"
                    for f, o in files
                )
            )
        )

    dask_dict = {}

    step_sum = 0
    for ttree in ttrees:
        entry_start = 0
        entry_stop = ttree.num_entries

        branchid_interpretation = {}
        for key in common_keys:
            branch = ttree[key]
            branchid_interpretation[branch.cache_key] = branch.interpretation
        ttree_step = _regularize_step_size(
            ttree, step_size, entry_start, entry_stop, branchid_interpretation
        )
        step_sum += int(ttree_step)

    entry_step = int(round(step_sum / len(ttrees)))
    assert entry_step >= 1

    for key in common_keys:
        dt = ttrees[0][key].interpretation.numpy_dtype
        if dt.subdtype is None:
            inner_shape = ()
        else:
            dt, inner_shape = dt.subdtype

        chunks = []
        chunk_args = []
        for i, ttree in enumerate(ttrees):
            entry_start = 0
            entry_stop = ttree.num_entries

            def foreach(start):
                stop = min(start + entry_step, entry_stop)  # noqa: B023
                length = stop - start
                chunks.append(length)  # noqa: B023
                chunk_args.append((i, start, stop))  # noqa: B023

            for start in range(entry_start, entry_stop, entry_step):
                foreach(start)

        if len(chunk_args) == 0:
            chunks.append(0)
            chunk_args.append((0, 0, 0))

        dask_dict[key] = _dask_array_from_map(
            _UprootReadNumpy(ttrees, key, interp_options),
            chunk_args,
            chunks=(tuple(chunks),),
            dtype=dt,
            label=f"{key}-from-uproot",
        )

    return dask_dict


def _get_dask_array_delay_open(
    files,
    filter_name=no_filter,
    filter_typename=no_filter,
    filter_branch=no_filter,
    recursive=True,
    full_paths=False,
    custom_classes=None,
    allow_missing=False,
    real_options=None,
    interp_options=None,
):
    ffile_path, fobject_path = files[0]
    obj = uproot._util.regularize_object_path(
        ffile_path, fobject_path, custom_classes, allow_missing, real_options
    )
    common_keys = obj.keys(
        recursive=recursive,
        filter_name=filter_name,
        filter_typename=filter_typename,
        filter_branch=filter_branch,
        full_paths=full_paths,
    )

    dask_dict = {}

    for key in common_keys:
        dt = obj[key].interpretation.numpy_dtype
        if dt.subdtype is None:
            inner_shape = ()
        else:
            dt, inner_shape = dt.subdtype

        dask_dict[key] = _dask_array_from_map(
            _UprootOpenAndReadNumpy(
                custom_classes, allow_missing, real_options, key, interp_options
            ),
            files,
            chunks=((numpy.nan,) * len(files),),
            dtype=dt,
            label=f"{key}-from-uproot",
        )
    return dask_dict


class _UprootRead:
    def __init__(self, ttrees, branches, interp_options) -> None:
        self.ttrees = ttrees
        self.branches = branches
        self.interp_options = interp_options

    def __call__(self, i_start_stop):
        i, start, stop = i_start_stop
        return self.ttrees[i].arrays(
            self.branches,
            entry_start=start,
            entry_stop=stop,
            ak_add_doc=self.interp_options["ak_add_doc"],
        )

    def project_columns(self, branches):
        if branches is not None:
            branches = [x for x in branches if x in self.branches]
        return _UprootRead(
            self.ttrees,
            branches,
            self.interp_options,
        )


class _UprootOpenAndRead:
    def __init__(
        self, custom_classes, allow_missing, real_options, common_keys, interp_options
    ) -> None:
        self.custom_classes = custom_classes
        self.allow_missing = allow_missing
        self.real_options = real_options
        self.common_keys = common_keys
        self.interp_options = interp_options

    def __call__(self, file_path_object_path):
        file_path, object_path = file_path_object_path
        ttree = uproot._util.regularize_object_path(
            file_path,
            object_path,
            self.custom_classes,
            self.allow_missing,
            self.real_options,
        )
        return ttree.arrays(
            self.common_keys, ak_add_doc=self.interp_options["ak_add_doc"]
        )

    def project_columns(self, common_keys):
        if common_keys is not None:
            common_keys = [x for x in common_keys if x in self.common_keys]

        return _UprootOpenAndRead(
            self.custom_classes,
            self.allow_missing,
            self.real_options,
            common_keys,
            self.interp_options,
        )


def _get_meta_array(
    awkward,
    dask_awkward,
    ttree,
    common_keys,
):
    form = awkward.forms.RecordForm(
        [ttree[key].interpretation.awkward_form(ttree.file) for key in common_keys],
        common_keys,
    )
    empty_arr = awkward.from_buffers(
        form, 0, {"": b"\x00\x00\x00\x00\x00\x00\x00\x00"}, buffer_key=""
    )
    return dask_awkward.core.typetracer_array(empty_arr)


def _get_dak_array(
    files,
    filter_name=no_filter,
    filter_typename=no_filter,
    filter_branch=no_filter,
    recursive=True,
    full_paths=False,
    step_size="100 MB",
    custom_classes=None,
    allow_missing=False,
    real_options=None,
    interp_options=None,
):
    dask_awkward = uproot.extras.dask_awkward()
    awkward = uproot.extras.awkward()

    ttrees = []
    common_keys = None
    is_self = []

    count = 0
    for file_path, object_path in files:
        obj = uproot._util.regularize_object_path(
            file_path, object_path, custom_classes, allow_missing, real_options
        )

        if obj is not None:
            count += 1

            if isinstance(obj, TBranch) and len(obj.keys(recursive=True)) == 0:
                original = obj
                obj = obj.parent
                is_self.append(True)

                def real_filter_branch(branch):
                    return branch is original and filter_branch(branch)  # noqa: B023

            else:
                is_self.append(False)
                real_filter_branch = filter_branch

            ttrees.append(obj)

            new_keys = obj.keys(
                recursive=recursive,
                filter_name=filter_name,
                filter_typename=filter_typename,
                filter_branch=real_filter_branch,
                full_paths=full_paths,
            )

            if common_keys is None:
                common_keys = new_keys
            else:
                new_keys = set(new_keys)
                common_keys = [key for key in common_keys if key in new_keys]

    if count == 0:
        raise ValueError(
            "allow_missing=True and no TTrees found in\n\n    {}".format(
                "\n    ".join(
                    "{"
                    + "{}: {}".format(
                        repr(f.file_path if isinstance(f, HasBranches) else f),
                        repr(f.object_path if isinstance(f, HasBranches) else o),
                    )
                    + "}"
                    for f, o in files
                )
            )
        )

    if len(common_keys) == 0 or not (all(is_self) or not any(is_self)):
        raise ValueError(
            "TTrees in\n\n    {}\n\nhave no TBranches in common".format(
                "\n    ".join(
                    "{"
                    + "{}: {}".format(
                        repr(f.file_path if isinstance(f, HasBranches) else f),
                        repr(f.object_path if isinstance(f, HasBranches) else o),
                    )
                    + "}"
                    for f, o in files
                )
            )
        )

    step_sum = 0
    for ttree in ttrees:
        entry_start = 0
        entry_stop = ttree.num_entries

        branchid_interpretation = {}
        for key in common_keys:
            branch = ttree[key]
            branchid_interpretation[branch.cache_key] = branch.interpretation
        ttree_step = _regularize_step_size(
            ttree, step_size, entry_start, entry_stop, branchid_interpretation
        )
        step_sum += int(ttree_step)

    entry_step = int(round(step_sum / len(ttrees)))

    partition_args = []
    for i, ttree in enumerate(ttrees):
        entry_start = 0
        entry_stop = ttree.num_entries

        def foreach(start):
            stop = min(start + entry_step, entry_stop)  # noqa: B023
            partition_args.append((i, start, stop))  # noqa: B023

        for start in range(entry_start, entry_stop, entry_step):
            foreach(start)

    meta = _get_meta_array(awkward, dask_awkward, ttrees[0], common_keys)

    if len(partition_args) == 0:
        partition_args.append((0, 0, 0))
    return dask_awkward.from_map(
        _UprootRead(ttrees, common_keys, interp_options),
        partition_args,
        label="from-uproot",
        meta=meta,
    )


def _get_dak_array_delay_open(
    files,
    filter_name=no_filter,
    filter_typename=no_filter,
    filter_branch=no_filter,
    recursive=True,
    full_paths=False,
    custom_classes=None,
    allow_missing=False,
    real_options=None,
    interp_options=None,
):
    dask_awkward = uproot.extras.dask_awkward()
    awkward = uproot.extras.awkward()

    ffile_path, fobject_path = files[0]
    obj = uproot._util.regularize_object_path(
        ffile_path, fobject_path, custom_classes, allow_missing, real_options
    )
    common_keys = obj.keys(
        recursive=recursive,
        filter_name=filter_name,
        filter_typename=filter_typename,
        filter_branch=filter_branch,
        full_paths=full_paths,
    )

    meta = _get_meta_array(awkward, dask_awkward, obj, common_keys)

    return dask_awkward.from_map(
        _UprootOpenAndRead(
            custom_classes, allow_missing, real_options, common_keys, interp_options
        ),
        files,
        label="from-uproot",
        meta=meta,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test_concatenate_numpy():
    files = skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root").replace(
        "6.20.04", "*"
    )
    arrays = uproot.concatenate({files: "sample"}, ["i8", "f8"], library="np")
    assert len(arrays["i8"]) == 420
    assert len(arrays["f8"]) == 420


def test_concatenate_awkward():
    awkward = pytest.importorskip("awkward")
    files = skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root").replace(
        "6.20.04", "*"
    )
    arrays = uproot.concatenate({files: "sample"}, ["i8", "f8"], library="ak")
    assert isinstance(arrays, awkward.Array)
    assert set(awkward.fields(arrays)) == {"i8", "f8"}
    assert len(arrays) == 420


def test_concatenate_pandas():
    pandas = pytest.importorskip("pandas")
    files = skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root").replace(
        "6.20.04", "*"
    )
    arrays = uproot.concatenate({files: "sample"}, ["i8", "f8"], library="pd")
    assert isinstance(arrays, pandas.DataFrame)
    assert set(arrays.columns.tolist()) == {"i8", "f8"}
    assert len(arrays) == 420



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest

import uproot


@pytest.mark.network
def test():
    with uproot.open(
        {"http://scikit-hep.org/uproot3/examples/HZZ.root": "events"}
    ) as t:
        t["MET_px"].array(library="np")
        t["MET_py"].array(library="np")



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines utilities for identifying the
:doc:`uproot.interpretation.Interpretation` of a
:doc:`uproot.behaviors.TBranch.TBranch`.

This includes a tokenizer/parser for C++ types and heuristics encoded in
:doc:`uproot.interpretation.identify.interpretation_of`. The latter will
need to be tweaked by new types, type combinations, and serialization methods
observed in ROOT files (perhaps forever), unless a systematic study can be
performed to exhaustively discover all cases.
"""


import ast
import re

import numpy

import uproot


def _normalize_ftype(fType):
    if fType is not None and uproot.const.kOffsetL < fType < uproot.const.kOffsetP:
        return fType - uproot.const.kOffsetL
    else:
        return fType


def _ftype_to_dtype(fType):
    fType = _normalize_ftype(fType)
    if fType == uproot.const.kBool:
        return numpy.dtype(numpy.bool_)
    elif fType == uproot.const.kChar:
        return numpy.dtype("i1")
    elif fType == uproot.const.kUChar:
        return numpy.dtype("u1")
    elif fType == uproot.const.kShort:
        return numpy.dtype(">i2")
    elif fType == uproot.const.kUShort:
        return numpy.dtype(">u2")
    elif fType == uproot.const.kInt:
        return numpy.dtype(">i4")
    elif fType in (uproot.const.kBits, uproot.const.kUInt, uproot.const.kCounter):
        return numpy.dtype(">u4")
    elif fType == uproot.const.kLong:
        return numpy.dtype(">i8")
    elif fType == uproot.const.kULong:
        return numpy.dtype(">u8")
    elif fType == uproot.const.kLong64:
        return numpy.dtype(">i8")
    elif fType == uproot.const.kULong64:
        return numpy.dtype(">u8")
    elif fType == uproot.const.kFloat:
        return numpy.dtype(">f4")
    elif fType == uproot.const.kDouble:
        return numpy.dtype(">f8")
    else:
        raise NotNumerical()


def _leaf_to_dtype(leaf, getdims):
    dims = ()
    if getdims:
        m = _title_has_dims.match(leaf.member("fTitle"))
        if m is not None:
            dims = tuple(eval(m.group(2).replace("][", ", ")))

    if leaf.classname == "TLeafO":
        return numpy.dtype((numpy.bool_, dims))
    elif leaf.classname == "TLeafB":
        if leaf.member("fIsUnsigned"):
            return numpy.dtype((numpy.uint8, dims))
        else:
            return numpy.dtype((numpy.int8, dims))
    elif leaf.classname == "TLeafS":
        if leaf.member("fIsUnsigned"):
            return numpy.dtype((numpy.uint16, dims))
        else:
            return numpy.dtype((numpy.int16, dims))
    elif leaf.classname == "TLeafI":
        if leaf.member("fIsUnsigned"):
            return numpy.dtype((numpy.uint32, dims))
        else:
            return numpy.dtype((numpy.int32, dims))
    elif leaf.classname == "TLeafL":
        if leaf.member("fIsUnsigned"):
            return numpy.dtype((numpy.uint64, dims))
        else:
            return numpy.dtype((numpy.int64, dims))
    elif leaf.classname == "TLeafF":
        return numpy.dtype((numpy.float32, dims))
    elif leaf.classname == "TLeafD":
        return numpy.dtype((numpy.float64, dims))
    elif leaf.classname == "TLeafElement":
        return numpy.dtype((_ftype_to_dtype(leaf.member("fType")), dims))
    else:
        raise NotNumerical()


_title_has_dims = re.compile(r"^([^\[\]]*)(\[[^\[\]]+\])+")
_item_dim_pattern = re.compile(r"\[([1-9][0-9]*)\]")
_item_any_pattern = re.compile(r"\[(.*)\]")


def _from_leaves_one(leaf, title):
    dims, is_jagged = (), False

    m = _title_has_dims.match(title)
    if m is not None:
        dims = tuple(int(x) for x in re.findall(_item_dim_pattern, title))
        if dims == ():
            if leaf.member("fLen") > 1:
                dims = (leaf.member("fLen"),)

        if any(
            _item_dim_pattern.match(x) is None
            for x in re.findall(_item_any_pattern, title)
        ):
            is_jagged = True

    return dims, is_jagged


def _from_leaves(branch, context):
    if len(branch.member("fLeaves")) == 0:
        raise UnknownInterpretation(
            "leaf-list with zero leaves",
            branch.file.file_path,
            branch.object_path,
        )

    elif len(branch.member("fLeaves")) == 1:
        leaf = branch.member("fLeaves")[0]
        title = leaf.member("fTitle")
        return _from_leaves_one(leaf, title)

    else:
        first = True
        for leaf in branch.member("fLeaves"):
            title = leaf.member("fTitle")
            if first:
                dims, is_jagged = _from_leaves_one(leaf, title)
            else:
                trial_dims, trial_is_jagged = _from_leaves_one(leaf, title)
                if dims != trial_dims or is_jagged != trial_is_jagged:
                    raise UnknownInterpretation(
                        "leaf-list with different dimensions among the leaves",
                        branch.file.file_path,
                        branch.object_path,
                    )
        return dims, is_jagged


def _float16_double32_walk_ast(node, branch, source):
    if isinstance(node, ast.AST):
        if (
            isinstance(node, ast.Name)
            and isinstance(node.ctx, ast.Load)
            and node.id.lower() == "pi"
        ):
            out = ast.Num(3.141592653589793)  # TMath::Pi()
        elif (
            isinstance(node, ast.Name)
            and isinstance(node.ctx, ast.Load)
            and node.id.lower() == "twopi"
        ):
            out = ast.Num(6.283185307179586)  # TMath::TwoPi()
        elif isinstance(node, ast.Num):
            out = ast.Num(float(node.n))
        elif isinstance(node, ast.BinOp) and isinstance(
            node.op, (ast.Add, ast.Sub, ast.Mult, ast.Div)
        ):
            out = ast.BinOp(
                _float16_double32_walk_ast(node.left, branch, source),
                node.op,
                _float16_double32_walk_ast(node.right, branch, source),
            )
        elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.USub):
            out = ast.UnaryOp(
                node.op, _float16_double32_walk_ast(node.operand, branch, source)
            )
        elif (
            isinstance(node, ast.List)
            and isinstance(node.ctx, ast.Load)
            and len(node.elts) == 2
        ):
            out = ast.List(
                [
                    _float16_double32_walk_ast(node.elts[0], branch, source),
                    _float16_double32_walk_ast(node.elts[1], branch, source),
                ],
                node.ctx,
            )
        elif (
            isinstance(node, ast.List)
            and isinstance(node.ctx, ast.Load)
            and len(node.elts) == 3
            and isinstance(node.elts[2], ast.Num)
        ):
            out = ast.List(
                [
                    _float16_double32_walk_ast(node.elts[0], branch, source),
                    _float16_double32_walk_ast(node.elts[1], branch, source),
                    node.elts[2],
                ],
                node.ctx,
            )
        else:
            raise UnknownInterpretation(
                f"cannot compute streamer title {source!r}",
                branch.file.file_path,
                branch.object_path,
            )
        out.lineno, out.col_offset = node.lineno, node.col_offset
        return out

    else:
        raise UnknownInterpretation(
            f"cannot compute streamer title {source!r}",
            branch.file.file_path,
            branch.object_path,
        )


def _float16_or_double32(branch, context, leaf, is_float16, dims):
    if leaf.classname in ("TLeafF16", "TLeafD32"):
        title = leaf.member("fTitle")
    elif branch.streamer is not None:
        title = branch.streamer.title
    else:
        title = ""

    try:
        left = title.index("[")
        right = title.index("]")

    except (ValueError, AttributeError):
        low, high, num_bits = 0, 0, "no brackets"  # distinct from "None"

    else:
        source = title[left : right + 1]

        try:
            parsed = ast.parse(source).body[0].value
            transformed = ast.Expression(
                _float16_double32_walk_ast(parsed, branch, source)
            )
            spec = eval(compile(transformed, repr(title), "eval"))
        except (UnknownInterpretation, SyntaxError):
            spec = ()

        if (
            len(spec) == 2
            and uproot._util.isnum(spec[0])
            and uproot._util.isnum(spec[1])
        ):
            low, high = spec
            num_bits = None

        elif (
            len(spec) == 3
            and uproot._util.isnum(spec[0])
            and uproot._util.isnum(spec[1])
            and uproot._util.isint(spec[2])
        ):
            low, high, num_bits = spec

        else:
            num_bits = "no brackets"

    if not is_float16:
        if num_bits == "no brackets":
            return uproot.interpretation.numerical.AsDtype(
                numpy.dtype((">f4", dims)), numpy.dtype(("f8", dims))
            )
        elif num_bits is None or not 2 <= num_bits <= 32:
            return uproot.interpretation.numerical.AsDouble32(low, high, 32, dims)
        else:
            return uproot.interpretation.numerical.AsDouble32(low, high, num_bits, dims)

    else:
        if num_bits == "no brackets":
            return uproot.interpretation.numerical.AsFloat16(low, high, 12, dims)
        elif num_bits is None or not 2 <= num_bits <= 32:
            return uproot.interpretation.numerical.AsFloat16(low, high, 32, dims)
        else:
            return uproot.interpretation.numerical.AsFloat16(low, high, num_bits, dims)


def interpretation_of(branch, context, simplify=True):
    """
    Args:
        branch (:doc:`uproot.behaviors.TBranch.TBranch`): The ``TBranch`` to
            interpret as an array.
        context (dict): Auxiliary data used in deserialization.
        simplify (bool): If True, call
            :ref:`uproot.interpretation.objects.AsObjects.simplify` on any
            :doc:`uproot.interpretation.objects.AsObjects` to try to get a
            more efficient interpretation.

    Attempts to derive an :doc:`uproot.interpretation.Interpretation` of the
    ``branch`` (within some ``context``).

    If no interpretation can be found, it raises
    :doc:`uproot.interpretation.identify.UnknownInterpretation`.
    """
    if len(branch.branches) != 0:
        if branch.top_level and branch.has_member("fClassName"):
            typename = str(branch.member("fClassName"))
        elif branch.streamer is not None:
            typename = str(branch.streamer.typename)
        else:
            typename = None
        subbranches = {x.name: x.interpretation for x in branch.branches}

        if typename == "TClonesArray":
            return uproot.interpretation.numerical.AsDtype(">i4")
        else:
            return uproot.interpretation.grouped.AsGrouped(
                branch, subbranches, typename=typename
            )

    if branch.classname == "TBranchObject":
        if branch.top_level and branch.has_member("fClassName"):
            model_cls = parse_typename(
                branch.member("fClassName"),
                file=branch.file,
                outer_header=True,
                inner_header=False,
                string_header=False,
            )
            return uproot.interpretation.objects.AsObjects(
                uproot.containers.AsDynamic(model_cls), branch
            )

        if branch.streamer is not None:
            model_cls = parse_typename(
                branch.streamer.typename,
                file=branch.file,
                outer_header=True,
                inner_header=False,
                string_header=True,
            )

            return uproot.interpretation.objects.AsObjects(
                uproot.containers.AsDynamic(model_cls), branch
            )

        return uproot.interpretation.objects.AsObjects(
            uproot.containers.AsDynamic(), branch
        )

    dims, is_jagged = _from_leaves(branch, context)

    try:
        if len(branch.member("fLeaves")) == 0:
            raise NotNumerical()

        elif len(branch.member("fLeaves")) == 1:
            leaf = branch.member("fLeaves")[0]

            leaftype = uproot.const.kBase
            if leaf.classname == "TLeafElement":
                leaftype = _normalize_ftype(leaf.member("fType"))

            is_float16 = (
                leaftype == uproot.const.kFloat16 or leaf.classname == "TLeafF16"
            )
            is_double32 = (
                leaftype == uproot.const.kDouble32 or leaf.classname == "TLeafD32"
            )

            if is_float16 or is_double32:
                out = _float16_or_double32(branch, context, leaf, is_float16, dims)

            else:
                from_dtype = _leaf_to_dtype(leaf, getdims=False).newbyteorder(">")

                if context.get("swap_bytes", True):
                    to_dtype = from_dtype.newbyteorder("=")
                else:
                    to_dtype = from_dtype

                out = uproot.interpretation.numerical.AsDtype(
                    numpy.dtype((from_dtype, dims)), numpy.dtype((to_dtype, dims))
                )

            if leaf.member("fLeafCount") is None:
                return out
            else:
                return uproot.interpretation.jagged.AsJagged(out)

        else:
            from_dtype = []
            for leaf in branch.member("fLeaves"):
                from_dtype.append(
                    (
                        leaf.member("fName"),
                        _leaf_to_dtype(leaf, getdims=True).newbyteorder(">"),
                    )
                )

            if context.get("swap_bytes", True):
                to_dtype = [(name, dt.newbyteorder("=")) for name, dt in from_dtype]
            else:
                to_dtype = from_dtype

            if all(
                leaf.member("fLeafCount") is None for leaf in branch.member("fLeaves")
            ):
                return uproot.interpretation.numerical.AsDtype(
                    numpy.dtype((from_dtype, dims)), numpy.dtype((to_dtype, dims))
                )
            else:
                raise UnknownInterpretation(
                    "leaf-list with non-null fLeafCount",
                    branch.file.file_path,
                    branch.object_path,
                )

    except NotNumerical:
        if (
            branch.member("fStreamerType", none_if_missing=True)
            == uproot.const.kTString
        ):
            return uproot.interpretation.strings.AsStrings(typename="TString")

        if len(branch.member("fLeaves")) == 1:
            leaf = branch.member("fLeaves")[0]

            if leaf.classname == "TLeafC":
                return uproot.interpretation.strings.AsStrings()

        elif len(branch.member("fLeaves")) > 1:
            raise UnknownInterpretation(
                "more than one TLeaf ({}) in a non-numerical TBranch".format(
                    len(branch.member("fLeaves"))
                ),
                branch.file.file_path,
                branch.object_path,
            ) from None

        if branch.top_level and branch.has_member("fClassName"):
            model_cls = parse_typename(
                branch.member("fClassName"),
                file=branch.file,
                outer_header=True,
                inner_header=False,
                string_header=False,
            )

            out = uproot.interpretation.objects.AsObjects(model_cls, branch)
            if simplify:
                return out.simplify()
            else:
                return out

        if branch.streamer is not None:
            model_cls = parse_typename(
                branch.streamer.typename,
                file=branch.file,
                outer_header=True,
                inner_header=False,
                string_header=True,
            )

            # kObjectp/kAnyp (as opposed to kObjectP/kAnyP) are stored inline
            if isinstance(
                model_cls, uproot.containers.AsPointer
            ) and branch.streamer.member("fType") in (
                uproot.const.kObjectp,
                uproot.const.kAnyp,
            ):
                while isinstance(model_cls, uproot.containers.AsPointer):
                    model_cls = model_cls.pointee

            if branch._streamer_isTClonesArray:
                if isinstance(branch.streamer, uproot.streamers.Model_TStreamerObject):
                    model_cls = uproot.containers.AsArray(False, False, model_cls, dims)
                else:
                    if hasattr(model_cls, "header"):
                        model_cls._header = False
                    model_cls = uproot.containers.AsArray(True, False, model_cls, dims)

            out = uproot.interpretation.objects.AsObjects(model_cls, branch)
            if simplify:
                return out.simplify()
            else:
                return out

    raise UnknownInterpretation(
        "none of the rules matched",
        branch.file.file_path,
        branch.object_path,
    )


_tokenize_typename_pattern = re.compile(
    r"(\b([A-Za-z_0-9]+)(\s*::\s*[A-Za-z_][A-Za-z_0-9]*)*\b(\s*\*)*|<|>|,)"
)

_simplify_token_1 = re.compile(r"\s*\*")
_simplify_token_2 = re.compile(r"\s*::\s*")
_simplify_token_3 = re.compile(r"\s*<\s*")
_simplify_token_4 = re.compile(r"\s*>\s*")


def _simplify_token(token, is_token=True):
    if is_token:
        text = token.group(0)
    else:
        text = token
    text = _simplify_token_1.sub("*", text)
    text = _simplify_token_2.sub("::", text)
    text = _simplify_token_3.sub("<", text)
    text = _simplify_token_4.sub(">", text)
    return text


def _parse_error(pos, typename, file):
    in_file = ""
    if file is not None:
        in_file = f"\nin file {file.file_path}"
    raise ValueError(
        """invalid C++ type name syntax at char {}

    {}
{}{}""".format(
            pos, typename, "-" * (4 + pos) + "^", in_file
        )
    )


def _parse_expect(what, tokens, i, typename, file):
    if i >= len(tokens):
        _parse_error(len(typename), typename, file)

    if what is not None and tokens[i].group(0) != what:
        _parse_error(tokens[i].start() + 1, typename, file)


def _parse_ignore_extra_arguments(tokens, i, typename, file, at_most):
    while tokens[i].group(0) == ",":
        if at_most == 0:
            _parse_error(tokens[i].start() + 1, typename, file)
        i, values = _parse_node(tokens, i + 1, typename, file, True, False, False)
        at_most -= 1

    return i


def _parse_maybe_quote(quoted, quote):
    if quote:
        return quoted
    else:
        return eval(quoted)


def _parse_node(tokens, i, typename, file, quote, header, inner_header):
    _parse_expect(None, tokens, i, typename, file)

    has2 = i + 1 < len(tokens)

    if tokens[i].group(0) == ",":
        _parse_error(tokens[i].start() + 1, typename, file)

    elif tokens[i].group(0) == "Bool_t":
        return i + 1, _parse_maybe_quote('numpy.dtype("?")', quote)
    elif tokens[i].group(0) == "bool":
        return i + 1, _parse_maybe_quote('numpy.dtype("?")', quote)

    elif _simplify_token(tokens[i]) == "Bool_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                f'uproot.containers.AsArray(False, {header}, numpy.dtype("?"))',
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "bool*":
        return (
            i + 1,
            _parse_maybe_quote(
                f'uproot.containers.AsArray(False, {header}, numpy.dtype("?"))',
                quote,
            ),
        )

    elif tokens[i].group(0) == "Char_t":
        return i + 1, _parse_maybe_quote('numpy.dtype("i1")', quote)
    elif tokens[i].group(0) == "char":
        return i + 1, _parse_maybe_quote('numpy.dtype("i1")', quote)
    elif tokens[i].group(0) == "UChar_t":
        return i + 1, _parse_maybe_quote('numpy.dtype("u1")', quote)
    elif has2 and tokens[i].group(0) == "unsigned" and tokens[i + 1].group(0) == "char":
        return i + 2, _parse_maybe_quote('numpy.dtype("u1")', quote)

    elif _simplify_token(tokens[i]) == "UChar_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype("u1"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif (
        has2
        and tokens[i].group(0) == "unsigned"
        and _simplify_token(tokens[i + 1]) == "char*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype("u1"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Short_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i2")', quote)
    elif tokens[i].group(0) == "short":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i2")', quote)
    elif tokens[i].group(0) == "UShort_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">u2")', quote)
    elif (
        has2 and tokens[i].group(0) == "unsigned" and tokens[i + 1].group(0) == "short"
    ):
        return i + 2, _parse_maybe_quote('numpy.dtype(">u2")', quote)

    elif _simplify_token(tokens[i]) == "Short_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i2"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "short*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i2"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "UShort_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u2"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif (
        has2
        and tokens[i].group(0) == "unsigned"
        and _simplify_token(tokens[i + 1]) == "short*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u2"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Int_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i4")', quote)
    elif tokens[i].group(0) == "int":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i4")', quote)
    elif tokens[i].group(0) == "UInt_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">u4")', quote)
    elif has2 and tokens[i].group(0) == "unsigned" and tokens[i + 1].group(0) == "int":
        return i + 2, _parse_maybe_quote('numpy.dtype(">u4")', quote)

    elif _simplify_token(tokens[i]) == "Int_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i4"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "int*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i4"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "UInt_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u4"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif (
        has2
        and tokens[i].group(0) == "unsigned"
        and _simplify_token(tokens[i + 1]) == "int*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u4"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif has2 and tokens[i].group(0) == tokens[i + 1].group(0) == "long":
        return i + 2, _parse_maybe_quote('numpy.dtype(">i8")', quote)
    elif (
        i + 2 < len(tokens)
        and tokens[i].group(0) == "unsigned"
        and tokens[i + 1].group(0) == tokens[i + 2].group(0) == "long"
    ):
        return i + 3, _parse_maybe_quote('numpy.dtype(">u8")', quote)

    elif tokens[i].group(0) == "Long_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i8")', quote)
    elif tokens[i].group(0) == "Long64_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i8")', quote)
    elif tokens[i].group(0) == "long":
        return i + 1, _parse_maybe_quote('numpy.dtype(">i8")', quote)
    elif tokens[i].group(0) == "ULong_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">u8")', quote)
    elif tokens[i].group(0) == "ULong64_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">u8")', quote)
    elif has2 and tokens[i].group(0) == "unsigned" and tokens[i + 1].group(0) == "long":
        return i + 2, _parse_maybe_quote('numpy.dtype(">u8")', quote)

    elif (
        has2
        and tokens[i].group(0) == "long"
        and _simplify_token(tokens[i + 1]) == "long*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                f'uproot.containers.AsArray({header}, numpy.dtype(">i8"))',
                quote,
            ),
        )
    elif (
        i + 2 < len(tokens)
        and tokens[i].group(0) == "unsigned"
        and _simplify_token(tokens[i + 1]) == "long"
        and _simplify_token(tokens[i + 2]) == "long*"
    ):
        return (
            i + 3,
            _parse_maybe_quote(
                f'uproot.containers.AsArray({header}, numpy.dtype(">u8"))',
                quote,
            ),
        )

    elif _simplify_token(tokens[i]) == "Long_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "Long64_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "long*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">i8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "ULong_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "ULong64_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif (
        has2
        and tokens[i].group(0) == "unsigned"
        and _simplify_token(tokens[i + 1]) == "long*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">u8"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Float_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">f4")', quote)
    elif tokens[i].group(0) == "float":
        return i + 1, _parse_maybe_quote('numpy.dtype(">f4")', quote)

    elif _simplify_token(tokens[i]) == "Float_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">f4"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "float*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">f4"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Double_t":
        return i + 1, _parse_maybe_quote('numpy.dtype(">f8")', quote)
    elif tokens[i].group(0) == "double":
        return i + 1, _parse_maybe_quote('numpy.dtype(">f8")', quote)

    elif _simplify_token(tokens[i]) == "Double_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">f8"))'.format(
                    header
                ),
                quote,
            ),
        )
    elif _simplify_token(tokens[i]) == "double*":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsArray(False, {}, numpy.dtype(">f8"))'.format(
                    header
                ),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Float16_t":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsFIXME("Float16_t in another context")', quote
            ),
        )

    elif _simplify_token(tokens[i]) == "Float16_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                "uproot.containers.AsArray(False, {}, "
                'uproot.containers.AsFIXME("Float16_t in array"))'.format(header),
                quote,
            ),
        )

    elif tokens[i].group(0) == "Double32_t":
        return (
            i + 1,
            _parse_maybe_quote(
                'uproot.containers.AsFIXME("Double32_t in another context")', quote
            ),
        )

    elif _simplify_token(tokens[i]) == "Double32_t*":
        return (
            i + 1,
            _parse_maybe_quote(
                "uproot.containers.AsArray(False, {}, "
                'uproot.containers.AsFIXME("Double32_t in array '
                '(note: Event.root fClosestDistance has an example)"))'.format(header),
                quote,
            ),
        )

    elif tokens[i].group(0) == "string" or _simplify_token(tokens[i]) == "std::string":
        return (
            i + 1,
            _parse_maybe_quote(f"uproot.containers.AsString({header})", quote),
        )
    elif tokens[i].group(0) == "TString":
        return (
            i + 1,
            _parse_maybe_quote(
                "uproot.containers.AsString(False, typename='TString')", quote
            ),
        )
    elif _simplify_token(tokens[i]) == "char*":
        return (
            i + 1,
            _parse_maybe_quote(
                "uproot.containers.AsString(False, length_bytes='4', typename='char*')",
                quote,
            ),
        )
    elif (
        has2
        and tokens[i].group(0) == "const"
        and _simplify_token(tokens[i + 1]) == "char*"
    ):
        return (
            i + 2,
            _parse_maybe_quote(
                "uproot.containers.AsString(False, length_bytes='4', typename='char*')",
                quote,
            ),
        )

    elif tokens[i].group(0) == "bitset" or _simplify_token(tokens[i]) == "std::bitset":
        _parse_expect("<", tokens, i + 1, typename, file)
        _parse_expect(None, tokens, i + 2, typename, file)
        try:
            num_bits = int(tokens[i + 2].group(0))
        except ValueError:
            _parse_error(tokens[i + 2].start() + 1, typename, file)
        # std::bitset only ever has one argument
        _parse_expect(">", tokens, i + 3, typename, file)
        return (
            i + 4,
            _parse_maybe_quote(
                f'uproot.containers.AsFIXME("std::bitset<{num_bits}>")',
                quote,
            ),
        )

    elif tokens[i].group(0) == "vector" or _simplify_token(tokens[i]) == "std::vector":
        _parse_expect("<", tokens, i + 1, typename, file)
        i, values = _parse_node(
            tokens, i + 2, typename, file, quote, inner_header, inner_header
        )
        i = _parse_ignore_extra_arguments(tokens, i, typename, file, 1)
        _parse_expect(">", tokens, i, typename, file)
        if quote:
            return (
                i + 1,
                f"uproot.containers.AsVector({header}, {values})",
            )
        else:
            return i + 1, uproot.containers.AsVector(header, values)

    elif (
        tokens[i].group(0) == "RVec"
        or _simplify_token(tokens[i]) == "VecOps::RVec"
        or _simplify_token(tokens[i]) == "ROOT::VecOps::RVec"
    ):
        _parse_expect("<", tokens, i + 1, typename, file)
        i, values = _parse_node(
            tokens, i + 2, typename, file, quote, inner_header, inner_header
        )
        i = _parse_ignore_extra_arguments(tokens, i, typename, file, 1)
        _parse_expect(">", tokens, i, typename, file)
        if quote:
            return (
                i + 1,
                f"uproot.containers.AsRVec({header}, {values})",
            )
        else:
            return i + 1, uproot.containers.AsRVec(header, values)

    elif tokens[i].group(0) == "set" or _simplify_token(tokens[i]) == "std::set":
        _parse_expect("<", tokens, i + 1, typename, file)
        i, keys = _parse_node(
            tokens, i + 2, typename, file, quote, inner_header, inner_header
        )
        i = _parse_ignore_extra_arguments(tokens, i, typename, file, 2)
        _parse_expect(">", tokens, i, typename, file)
        if quote:
            return i + 1, f"uproot.containers.AsSet({header}, {keys})"
        else:
            return i + 1, uproot.containers.AsSet(header, keys)

    elif tokens[i].group(0) == "map" or _simplify_token(tokens[i]) == "std::map":
        _parse_expect("<", tokens, i + 1, typename, file)
        i, keys = _parse_node(
            tokens, i + 2, typename, file, quote, header, inner_header
        )
        _parse_expect(",", tokens, i, typename, file)
        i, values = _parse_node(
            tokens, i + 1, typename, file, quote, header, inner_header
        )
        i = _parse_ignore_extra_arguments(tokens, i, typename, file, 2)
        _parse_expect(">", tokens, i, typename, file)
        if quote:
            return (
                i + 1,
                f"uproot.containers.AsMap({header}, {keys}, {values})",
            )
        else:
            return i + 1, uproot.containers.AsMap(header, keys, values)

    else:
        start, stop = tokens[i].span()

        if has2 and tokens[i + 1].group(0) == "<":
            i, keys = _parse_node(
                tokens, i + 2, typename, file, quote, inner_header, inner_header
            )
            while tokens[i].group(0) == ",":
                i, keys = _parse_node(
                    tokens, i + 1, typename, file, quote, inner_header, inner_header
                )
            _parse_expect(">", tokens, i, typename, file)
            stop = tokens[i].span()[1]

        classname = _simplify_token(typename[start:stop], is_token=False)
        classname = uproot.model.classname_regularize(classname)

        pointers = 0
        while classname.endswith("*"):
            pointers += 1
            classname = classname[:-1]

        if quote:
            cls = f"c({classname!r})"
            for _ in range(pointers):
                cls = f"uproot.containers.AsPointer({cls})"
        elif file is None:
            cls = uproot.classes[classname]
            for _ in range(pointers):
                cls = uproot.containers.AsPointer(cls)
        else:
            cls = file.class_named(classname)
            for _ in range(pointers):
                cls = uproot.containers.AsPointer(cls)

        return i + 1, cls


def parse_typename(
    typename,
    file=None,
    quote=False,
    outer_header=True,
    inner_header=False,
    string_header=False,
):
    """
    Args:
        typename (str): The C++ type to parse.
        file (None or :doc:`uproot.reading.CommonFileMethods`): Used to provide
            error messages with the ``file_path``.
        quote (bool): If True, return the output as a string to evaluate. This
            is used to build code for a :doc:`uproot.model.Model`, rather than
            the :doc:`uproot.model.Model` itself.
        outer_header (bool): If True, set the ``header`` flag for the outermost
            :doc:`uproot.containers.AsContainer` to True.
        inner_header (bool): If True, set the ``header`` flag for inner
            :doc:`uproot.containers.AsContainer` objects to True.
        string_header (bool): If True, set the ``header`` flag for
            :doc:`uproot.containers.AsString` objects to True.

    Return a :doc:`uproot.model.Model` or :doc:`uproot.containers.AsContainer`
    for the C++ ``typename``.
    """
    tokens = list(_tokenize_typename_pattern.finditer(typename))

    if (
        not string_header
        and len(tokens) != 0
        and (
            tokens[0].group(0) == "string"
            or _simplify_token(tokens[0]) == "std::string"
        )
    ):
        i, out = 1, _parse_maybe_quote("uproot.containers.AsString(False)", quote)

    else:
        i, out = _parse_node(
            tokens, 0, typename, file, quote, outer_header, inner_header
        )

    if i < len(tokens):
        _parse_error(tokens[i].start(), typename, file)

    return out


class NotNumerical(Exception):
    """
    Exception used to stop searches for a numerical interpretation in
    :doc:`uproot.interpretation.identify.interpretation_of` as soon as a
    non-conforming type is found.
    """

    pass


class UnknownInterpretation(Exception):
    """
    Exception raised by :doc:`uproot.interpretation.identify.interpretation_of`
    if an :doc:`uproot.interpretation.Interpretation` cannot be found.

    The :ref:`uproot.behaviors.TBranch.TBranch.interpretation` property may have
    :doc:`uproot.interpretation.identify.UnknownInterpretation` as a value.

    Any attempts to use this class as a
    :doc:`uproot.interpretation.Interpretation` causes it to raise itself.
    Thus, failing to find an interpretation for a ``TBranch`` is not a fatal
    error, but attempting to use it to deserialize arrays is a fatal error.
    """

    def __init__(self, reason, file_path, object_path):
        self.reason = reason
        self.file_path = file_path
        self.object_path = object_path

    def __repr__(self):
        return f"<UnknownInterpretation {self.reason!r}>"

    def __str__(self):
        return """{}
in file {}
in object {}""".format(
            self.reason, self.file_path, self.object_path
        )

    @property
    def typename(self):
        return "unknown"

    @property
    def cache_key(self):
        raise self

    @property
    def numpy_dtype(self):
        raise self

    def awkward_form(
        self,
        file,
        context=None,
        index_format="i64",
        header=False,
        tobject_header=False,
        breadcrumbs=(),
    ):
        raise self

    @property
    def basket_array(self):
        raise self

    @property
    def final_array(self):
        raise self

    @property
    def hook_before_basket_array(self):
        raise self

    @property
    def hook_after_basket_array(self):
        raise self

    @property
    def hook_before_final_array(self):
        raise self

    @property
    def hook_before_library_finalize(self):
        raise self

    @property
    def hook_after_final_array(self):
        raise self

    @property
    def itemsize(self):
        raise self

    @property
    def from_dtype(self):
        raise self

    @property
    def to_dtype(self):
        raise self

    @property
    def content(self):
        raise self

    @property
    def header_bytes(self):
        raise self

    @property
    def size_1to5_bytes(self):
        raise self



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test_numpy():
    with uproot.open(skhep_testdata.data_path("uproot-issue-643.root"))[
        "ntuple0/objects"
    ] as t:
        array = t["pf"].array(entry_stop=2, library="np")
        assert array[1][1].member("first").member("fE") == 326.0029230449897
        assert array[1][1].member("second") == 22


awkward = pytest.importorskip("awkward")


def test_awkward():
    with uproot.open(skhep_testdata.data_path("uproot-issue-643.root"))[
        "ntuple0/objects"
    ] as t:
        array = t["pf"].array(entry_stop=2, library="ak")
        assert array[1, 1, "first", "fE"] == 326.0029230449897
        assert array[1, 1, "second"] == 22



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import pytest

import uproot._util
import uproot.reading


@pytest.mark.skipif(
    not uproot._util.win, reason="Drive letters only parsed on Windows."
)
def test_windows_drive_letters():
    assert (
        uproot._util.file_path_to_source_class(
            "file:///g:/mydir/file.root", uproot.reading.open.defaults
        )[1]
        == "g:/mydir/file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            "file:/g:/mydir/file.root", uproot.reading.open.defaults
        )[1]
        == "g:/mydir/file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            "file:g:/mydir/file.root", uproot.reading.open.defaults
        )[1]
        == "g:/mydir/file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            "/g:/mydir/file.root", uproot.reading.open.defaults
        )[1]
        == "g:/mydir/file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            r"\g:/mydir/file.root", uproot.reading.open.defaults
        )[1]
        == "g:/mydir/file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            r"g:\mydir\file.root", uproot.reading.open.defaults
        )[1]
        == r"g:\mydir\file.root"
    )

    assert (
        uproot._util.file_path_to_source_class(
            r"\g:\mydir\file.root", uproot.reading.open.defaults
        )[1]
        == r"g:\mydir\file.root"
    )


def test_escaped_uri_codes():
    # If they're file:// paths, yes we should unquote the % signs.
    assert (
        uproot._util.file_path_to_source_class(
            "file:///my%20file.root", uproot.reading.open.defaults
        )[1]
        == "/my file.root"
    )
    assert (
        uproot._util.file_path_to_source_class(
            "file:///my%E2%80%92file.root", uproot.reading.open.defaults
        )[1]
        == "/my\u2012file.root"
    )

    # Otherwise, no we should not.
    assert (
        uproot._util.file_path_to_source_class(
            "/my%20file.root", uproot.reading.open.defaults
        )[1]
        == "/my%20file.root"
    )
    assert (
        uproot._util.file_path_to_source_class(
            "/my%E2%80%92file.root", uproot.reading.open.defaults
        )[1]
        == "/my%E2%80%92file.root"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os
import sys

import numpy
import pytest

import uproot


def tobytes(x):
    if hasattr(x, "tobytes"):
        return x.tobytes()
    else:
        return x.tostring()


def test_file(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")
    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    expected = [
        b"******",
        b"    ",
        b"...",
        b"+++++++",
        b"!!!!!",
        b"@@@@@",
    ]

    for num_workers in [1, 2]:
        with uproot.source.file.MultithreadedFileSource(
            filename, num_workers=num_workers
        ) as source:
            for i, (start, stop) in enumerate(
                [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)]
            ):
                chunk = source.chunk(start, stop)
                assert tobytes(chunk.raw_data) == expected[i]

        with pytest.raises(Exception):
            uproot.source.file.MultithreadedFileSource(
                filename + "-does-not-exist", num_workers=num_workers
            )


def test_memmap(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")
    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    expected = [
        b"******",
        b"    ",
        b"...",
        b"+++++++",
        b"!!!!!",
        b"@@@@@",
    ]

    with uproot.source.file.MemmapSource(filename, num_fallback_workers=1) as source:
        for i, (start, stop) in enumerate(
            [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)]
        ):
            chunk = source.chunk(start, stop)
            assert tobytes(chunk.raw_data) == expected[i]

    with pytest.raises(Exception):
        uproot.source.file.MemmapSource(
            filename + "-does-not-exist", num_fallback_workers=1
        )


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http():
    for num_workers in [1, 2]:
        with uproot.source.http.MultithreadedHTTPSource(
            "https://example.com", num_workers=num_workers, timeout=10
        ) as source:
            for start, stop in [(0, 100), (50, 55), (200, 400)]:
                chunk = source.chunk(start, stop)
                assert len(tobytes(chunk.raw_data)) == stop - start


@pytest.mark.network
def test_http_fail():
    for num_workers in [1, 2]:
        with pytest.raises(Exception):
            with uproot.source.http.MultithreadedHTTPSource(
                "https://wonky.cern/does-not-exist",
                num_workers=num_workers,
                timeout=0.1,
            ) as source:
                source.chunk(0, 100)


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http_multipart():
    with uproot.source.http.HTTPSource(
        "https://example.com", timeout=10, num_fallback_workers=1
    ) as source:
        for start, stop in [(0, 100), (50, 55), (200, 400)]:
            chunk = source.chunk(start, stop)
            assert len(tobytes(chunk.raw_data)) == stop - start


@pytest.mark.network
def test_http_multipart_fail():
    with pytest.raises(Exception):
        with uproot.source.http.HTTPSource(
            "https://wonky.cern/does-not-exist", timeout=0.1, num_fallback_workers=1
        ) as source:
            tobytes(source.chunk(0, 100).raw_data)


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        num_workers=1,
        timeout=10,
    ) as source:
        one = tobytes(source.chunk(0, 100).raw_data)
        assert len(one) == 100
        two = tobytes(source.chunk(50, 55).raw_data)
        assert len(two) == 5
        three = tobytes(source.chunk(200, 400).raw_data)
        assert len(three) == 200
        assert one[:4] == b"root"


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_worker():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        num_workers=5,
        timeout=10,
    ) as source:
        one = tobytes(source.chunk(0, 100).raw_data)
        assert len(one) == 100
        two = tobytes(source.chunk(50, 55).raw_data)
        assert len(two) == 5
        three = tobytes(source.chunk(200, 400).raw_data)
        assert len(three) == 200
        assert one[:4] == b"root"


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        one = tobytes(source.chunk(0, 100).raw_data)
        assert len(one) == 100
        two = tobytes(source.chunk(50, 55).raw_data)
        assert len(two) == 5
        three = tobytes(source.chunk(200, 400).raw_data)
        assert len(three) == 200
        assert one[:4] == b"root"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test_deep_mkdir(tmp_path):
    filename = os.path.join(tmp_path, "whatever.root")

    with uproot.recreate(filename) as f1:
        f1.mkdir("/one/two/three")

    assert uproot.open(filename).keys() == ["one;1", "one/two;1", "one/two/three;1"]


def test_bulk_copy(tmp_path):
    source_filename = os.path.join(tmp_path, "source.root")
    dest_filename = os.path.join(tmp_path, "dest.root")

    f_histogram = ROOT.TFile(source_filename, "recreate")
    f_histogram.mkdir("subdir")
    f_histogram.cd("subdir")
    x = ROOT.TH1F("hist", "title", 100, -5, 5)
    x.Write()
    f_histogram.Close()

    with uproot.open(source_filename) as source:
        with uproot.recreate(dest_filename) as dest:
            dest.copy_from(source, filter_name="subdir/hist")

    with uproot.open(dest_filename) as dest:
        assert dest.keys() == ["subdir;1", "subdir/hist;1"]
        hist = dest["subdir/hist"]
        assert hist.name == "hist"
        assert hist.title == "title"
        assert hist.axis().low == -5

    f2 = ROOT.TFile(dest_filename, "update")
    h2 = f2.Get("subdir/hist")
    assert h2.GetName() == "hist"
    assert h2.GetTitle() == "title"
    assert h2.GetNbinsX() == 100

    y = ROOT.TObjString("hello")
    y.Write()
    f2.Close()

    assert set(uproot.open(dest_filename).keys()) == {
        "subdir;1",
        "subdir/hist;1",
        "hello;1",
    }



"""
Use this to make src/uproot/models/TH.py and src/uproot/models/TGraph.py.
"""

import numpy as np

import uproot

keys = [
    "th1c",
    "th1d",
    "th1f",
    "th1i",
    "th1s",
    "th2c",
    "th2d",
    "th2f",
    "th2i",
    "th2s",
    "th3c",
    "th3d",
    "th3f",
    "th3i",
    "th3s",
    "tprofile",
    "tprofile2d",
    "tprofile3d",
    "tgraphasymmerrors",
    "tgrapherrors",
    "tgraph",
]

superclasses = [
    # ("TCollection", 3),
    # ("TSeqCollection", 0),
    # ("TList", 5),
    # ("THashList", 0),
    ("TAttAxis", 4),
    ("TAxis", 10),
    # ("TAttMarker", 2),
    # ("TAttFill", 2),
    # ("TAttLine", 2),
    # ("TString", 2),
    # ("TObject", 1),
    # ("TNamed", 1),
    ("TH1", 8),
    ("TH2", 5),
    ("TAtt3D", 1),
    ("TH3", 6),
]


# with uproot.open("example-objects.root") as f:
#     f.file.streamers
#     pairs = []
#     for key in keys:
#         obj = f[key]
#         for pair in f.file.streamer_dependencies(obj.classname, obj.class_version):
#             if pair not in pairs and pair[0].lower() not in keys:
#                 pairs.append(pair)
#     for streamer_name, streamer_version in pairs:
#         print(streamer_name, streamer_version)


with uproot.open("example-objects.root") as f:
    f.file.streamers

    for classname, class_version in superclasses:
        cls = f.file.class_named(classname, class_version)
        print(cls.class_code)
        print(
            f"""
    writable = True

    def _serialize(self, out, header, name):
        where = len(out)
        for x in self._bases:
            x._serialize(out, True, name)
        raise NotImplementedError("FIXME")
        if header:
            num_bytes = sum(len(x) for x in out[where:])
            version = {class_version}
            out.insert(where, uproot.serialization.numbytes_version(num_bytes, version))

class {uproot.model.classname_encode(classname)}(uproot.model.DispatchByVersion):
    \"\"\"
    A :doc:`uproot.model.DispatchByVersion` for ``{classname}``.
    \"\"\"

    known_versions = {{{class_version}: {cls.__name__}}}
"""
        )

    for key in keys:
        obj = f[key]
        print(type(obj).class_code)
        print(
            """
    class_rawstreamers = ("""
        )
        for streamer_name, streamer_version in f.file.streamer_dependencies(
            obj.classname, obj.class_version
        ):
            streamer = f.file.streamer_named(streamer_name, streamer_version)
            inner = streamer.serialize()
            header = np.array(
                [(len(inner) + 18) | uproot.const.kByteCountMask], ">u4"
            ).tobytes()
            preamble = b"\xff\xff\xff\xffTStreamerInfo\x00"
            full = header + preamble + inner + b"\x00"
            print(
                f"""        uproot._writing.RawStreamerInfo(
            None,
            {full},
            {streamer_name!r},
            {streamer_version},
        ),"""
            )
        print(
            f"""    )
    writable = True

    def _serialize(self, out, header, name):
        where = len(out)
        for x in self._bases:
            x._serialize(out, True, name)
        raise NotImplementedError("FIXME")
        if header:
            num_bytes = sum(len(x) for x in out[where:])
            version = {obj.class_version}
            out.insert(where, uproot.serialization.numbytes_version(num_bytes, version))
"""
        )

        print(
            f"""
class {uproot.model.classname_encode(obj.classname)}(uproot.model.DispatchByVersion):
    \"\"\"
    A :doc:`uproot.model.DispatchByVersion` for ``{obj.classname}``.
    \"\"\"

    known_versions = {{{obj.class_version}: {type(obj).__name__}}}
"""
        )

    for classname, _ in superclasses:
        print(
            f"uproot.classes[{classname!r}] = {uproot.model.classname_encode(classname)}"
        )

    for key in keys:
        obj = f[key]
        print(
            f"uproot.classes[{obj.classname!r}] = {uproot.model.classname_encode(obj.classname)}"
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import pytest
import skhep_testdata

import uproot


def test():
    for arrays in uproot.iterate(
        skhep_testdata.data_path("uproot-issue335.root") + ":empty_tree",
        ["var"],
        library="np",
    ):
        pass



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test_volley(tmp_path):
    filename = os.path.join(tmp_path, "testy.root")

    f1 = ROOT.TFile(filename, "recreate")
    f1.Close()

    assert set(uproot.open(filename).file.streamers) == set()

    with uproot.writing.update(filename) as f2:
        f2.file._cascading.streamers.write(f2.file.sink)

    assert set(uproot.open(filename).file.streamers) == set()

    f3 = ROOT.TFile(filename, "update")
    x = ROOT.TObjString("hello")
    x.Write()
    f3.Close()

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    with uproot.writing.update(filename) as f4:
        f4.file._cascading.streamers.write(f4.file.sink)

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    f5 = ROOT.TFile(filename, "update")
    y = ROOT.TH1F("hey", "there", 100, -5, 5)
    y.Write()
    f5.Close()

    with uproot.writing.update(filename) as f6:
        f6.file._cascading.streamers.write(f6.file.sink)


def test_with_mkdir(tmp_path):
    filename = os.path.join(tmp_path, "testy.root")

    f1 = ROOT.TFile(filename, "recreate")
    f1.Close()

    assert set(uproot.open(filename).file.streamers) == set()

    with uproot.writing.update(filename) as f2:
        f2.mkdir("one")

    assert set(uproot.open(filename).file.streamers) == set()

    f3 = ROOT.TFile(filename, "update")
    f3.cd("one")
    x = ROOT.TObjString("hello")
    x.Write()
    f3.Close()

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    with uproot.writing.update(filename) as f4:
        f4.mkdir("two")

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    f5 = ROOT.TFile(filename, "update")
    y = ROOT.TH1F("hey", "there", 100, -5, 5)
    y.Write()
    f5.Close()

    with uproot.writing.update(filename) as f6:
        f6.mkdir("three")


def test_add_streamers1(tmp_path):
    has_TObjString = os.path.join(tmp_path, "has_TObjString.root")

    f_TObjString = ROOT.TFile(has_TObjString, "recreate")
    x = ROOT.TObjString("hello")
    x.Write()
    f_TObjString.Close()

    streamers = [uproot.open(has_TObjString).file.streamers["TObjString"][1]]

    filename = os.path.join(tmp_path, "testy.root")

    f1 = ROOT.TFile(filename, "recreate")
    f1.Close()

    assert set(uproot.open(filename).file.streamers) == set()

    with uproot.writing.update(filename) as f2:
        f2.file.update_streamers(streamers)

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    f3 = ROOT.TFile(filename, "update")
    y = ROOT.TObjString("there")
    y.Write()
    f3.Close()

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    with uproot.writing.update(filename) as f3:
        pass

    assert set(uproot.open(filename).file.streamers) == {"TObjString"}

    f4 = ROOT.TFile(filename, "update")
    z = ROOT.TH1F("histogram", "", 100, -5, 5)
    z.Write()
    f4.Close()


def test_add_streamers2(tmp_path):
    has_histogram = os.path.join(tmp_path, "has_histogram.root")

    f_histogram = ROOT.TFile(has_histogram, "recreate")
    x = ROOT.TH1F("histogram", "", 100, -5, 5)
    x.Write()
    f_histogram.Close()

    streamers = [
        byversion
        for byname in uproot.open(has_histogram).file.streamers.values()
        for byversion in byname.values()
    ]

    filename = os.path.join(tmp_path, "testy.root")

    f1 = ROOT.TFile(filename, "recreate")
    f1.Close()

    assert set(uproot.open(filename).file.streamers) == set()

    with uproot.writing.update(filename) as f2:
        f2.file.update_streamers(streamers)

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
    }

    f3 = ROOT.TFile(filename, "update")
    y = ROOT.TH1F("margotsih", "", 100, -5, 5)
    f3.Close()

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
    }

    with uproot.writing.update(filename) as f3:
        pass

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
    }

    f3 = ROOT.TFile(filename, "update")
    z = ROOT.TObjString("there")
    z.Write()
    f3.Close()


def test_add_streamers3(tmp_path):
    has_TObjString = os.path.join(tmp_path, "has_TObjString.root")

    f_TObjString = ROOT.TFile(has_TObjString, "recreate")
    x = ROOT.TObjString("hello")
    x.Write()
    f_TObjString.Close()

    streamers = [uproot.open(has_TObjString).file.streamers["TObjString"][1]]

    filename = os.path.join(tmp_path, "testy.root")

    f1 = ROOT.TFile(filename, "recreate")
    y = ROOT.TH1F("histogram", "", 100, -5, 5)
    y.Write()
    f1.Close()

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
    }

    with uproot.writing.update(filename) as f2:
        f2.file.update_streamers(streamers)

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
        "TObjString",
    }

    f2 = ROOT.TFile(filename, "update")
    assert {z.GetName() for z in f2.GetStreamerInfoList()} == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
        "TObjString",
    }
    w = ROOT.TObjString("wow")
    w.Write()
    f2.Close()

    assert set(uproot.open(filename).file.streamers) == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
        "TObjString",
    }

    f3 = ROOT.TFile(filename, "update")
    assert {z.GetName() for z in f3.GetStreamerInfoList()} == {
        "TObject",
        "TNamed",
        "TH1F",
        "TH1",
        "TAttLine",
        "TAttFill",
        "TAttMarker",
        "TAxis",
        "TAttAxis",
        "THashList",
        "TList",
        "TSeqCollection",
        "TCollection",
        "TString",
        "TObjString",
    }
    f3.Close()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import queue
import sys

import awkward as ak
import numpy
import pytest
import skhep_testdata

import uproot


def test_rntuple_stl_containers():
    filename = skhep_testdata.data_path("test_ntuple_stl_containers.root")
    with uproot.open(filename) as f:
        R = f["ntuple"]
        assert R.keys() == [
            "string",
            "vector_int32",
            "vector_vector_int32",
            "vector_string",
            "vector_vector_string",
            "variant_int32_string",
            "vector_variant_int64_string",
            "tuple_int32_string",
            "vector_tuple_int32_string",
        ]
        r = R.arrays()
        assert ak.all(r["string"] == ["one", "two"])

        assert r["vector_int32"][0] == [1]
        assert ak.all(r["vector_int32"][1] == [1, 2])
        assert r["vector_vector_int32"][0] == [[1]]

        assert ak.all(r["vector_vector_int32"][1] == [[1], [2]])

        assert r["vector_string"][0] == ["one"]
        assert ak.all(r["vector_string"][1] == ["one", "two"])

        assert ak.all(r["vector_vector_string"] == [["one"], [["one"], ["two"]]])
        assert r["variant_int32_string"][0] == 1
        assert r["variant_int32_string"][1] == "two"

        assert r["vector_variant_int64_string"][0][0] == "one"
        assert r["vector_variant_int64_string"][1][0] == "one"
        assert r["vector_variant_int64_string"][1][1] == 2
        assert type(r["vector_variant_int64_string"][1][1]) == numpy.int64

        assert ak.all(r["tuple_int32_string"]._0 == [1, 2])
        assert ak.all(r["tuple_int32_string"]._1 == ["one", "two"])
        assert list(r["tuple_int32_string"][0].to_list().values()) == [1, "one"]

        assert r["vector_tuple_int32_string"][0]._0 == [1]
        assert r["vector_tuple_int32_string"][0]._1 == ["one"]
        assert ak.all(r["vector_tuple_int32_string"][1]._0 == [1, 2])
        assert ak.all(r["vector_tuple_int32_string"][1]._1 == ["one", "two"])
        assert ak.all(r["vector_tuple_int32_string"]._0 == [[1], [1, 2]])
        assert ak.all(r["vector_tuple_int32_string"]._1 == [["one"], ["one", "two"]])



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines integer constants used by serialization and deserialization routines.
"""


import numpy

# determines when a file is "big"
kStartBigFile = 2000000000

# used in unmarshaling
kByteCountMask = numpy.int64(0x40000000)
kByteCountVMask = numpy.int64(0x4000)
kClassMask = numpy.int64(0x80000000)
kNewClassTag = numpy.int64(0xFFFFFFFF)

kIsOnHeap = numpy.uint32(0x01000000)
kNotDeleted = numpy.uint32(0x02000000)
kMustCleanup = numpy.uint32(1 << 3)
kIsReferenced = numpy.uint32(1 << 4)

kMapOffset = 2

# not used?
kNullTag = 0
kZombie = numpy.uint32(0x04000000)
kBitMask = numpy.uint32(0x00FFFFFF)
kDisplacementMask = numpy.uint32(0xFF000000)

############# core/zip/inc/Compression.h

kZLIB = 1
kLZMA = 2
kOldCompressionAlgo = 3
kLZ4 = 4
kZSTD = 5
kUndefinedCompressionAlgorithm = 6

############# constants for streamers

kBase = 0
kChar = 1
kShort = 2
kInt = 3
kLong = 4
kFloat = 5
kCounter = 6
kCharStar = 7
kDouble = 8
kDouble32 = 9
kLegacyChar = 10
kUChar = 11
kUShort = 12
kUInt = 13
kULong = 14
kBits = 15
kLong64 = 16
kULong64 = 17
kBool = 18
kFloat16 = 19
kOffsetL = 20
kOffsetP = 40
kObject = 61
kAny = 62
kObjectp = 63
kObjectP = 64
kTString = 65
kTObject = 66
kTNamed = 67
kAnyp = 68
kAnyP = 69
kAnyPnoVT = 70
kSTLp = 71

kSkip = 100
kSkipL = 120
kSkipP = 140

kConv = 200
kConvL = 220
kConvP = 240

kSTL = 300
kSTLstring = 365

kStreamer = 500
kStreamLoop = 501

############# constants from core/foundation/inc/ESTLType.h

kNotSTL = 0
kSTLvector = 1
kSTLlist = 2
kSTLdeque = 3
kSTLmap = 4
kSTLmultimap = 5
kSTLset = 6
kSTLmultiset = 7
kSTLbitset = 8
kSTLforwardlist = 9
kSTLunorderedset = 10
kSTLunorderedmultiset = 11
kSTLunorderedmap = 12
kSTLunorderedmultimap = 13
kSTLend = 14
kSTLany = 300

############# IOFeatures

kGenerateOffsetMap = numpy.uint8(1)

############# other

kStreamedMemberWise = numpy.uint16(1 << 14)

############ RNTuple https://github.com/root-project/root/blob/master/tree/ntuple/v7/doc/specifications.md
rntuple_col_num_to_dtype_dict = {
    1: "uint64",
    2: "uint32",
    3: "uint64",  # Switch
    4: "uint8",
    5: "uint8",  # char
    6: "bit",
    7: "float64",
    8: "float32",
    9: "float16",
    10: "int64",
    11: "int32",
    12: "int16",
    13: "int8",
    14: "uint32",  # SplitIndex64 delta encoding
    15: "uint64",  # SplitIndex32 delta encoding
    16: "float64",  # split
    17: "float32",  # split
    18: "float16",  # split
    19: "int64",  # split
    20: "int32",  # split
    21: "int16",  # split
}

rntuple_col_type_to_num_dict = {
    "index64": 1,
    "index32": 2,
    "switch": 3,
    "byte": 4,
    "char": 5,
    "bit": 6,
    "real64": 7,
    "real32": 8,
    "real16": 9,
    "int64": 10,
    "int32": 11,
    "int16": 12,
    "int8": 13,
    "splitindex64": 14,
    "splitindex32": 15,
    "splitreal64": 16,
    "splitreal32": 17,
    "splitreal16": 18,
    "splitin64": 19,
    "splitint32": 20,
    "splitint16": 21,
}

rntuple_role_leaf = 0
rntuple_role_vector = 1
rntuple_role_struct = 2
rntuple_role_union = 3



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import queue
import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test_flat():
    filename = skhep_testdata.data_path("test_ntuple_int_float.root")
    with uproot.open(filename) as f:
        R = f["ntuple"]
        assert R.keys() == ["one_integers", "two_floats"]
        assert [r.type_name for r in R.header.field_records] == [
            "std::int32_t",
            "float",
        ]
        assert R.header.crc32 == R.footer.header_crc32
        assert all(R.arrays(entry_stop=3)["one_integers"] == numpy.array([9, 8, 7]))
        assert all(
            R.arrays("one_integers", entry_stop=3)["one_integers"]
            == numpy.array([9, 8, 7])
        )
        assert all(
            R.arrays(entry_start=1, entry_stop=3)["one_integers"] == numpy.array([8, 7])
        )

    filename = skhep_testdata.data_path("test_ntuple_int_5e4.root")
    with uproot.open(filename) as f:
        R = f["ntuple"]
        assert all(
            R.arrays(entry_stop=3)["one_integers"] == numpy.array([50000, 49999, 49998])
        )
        assert all(R.arrays(entry_start=-3)["one_integers"] == numpy.array([3, 2, 1]))


def test_jagged():
    filename = skhep_testdata.data_path("test_ntuple_int_vfloat_tlv_vtlv.root")
    with uproot.open(filename) as f:
        R = f["ntuple"]
        assert R.keys() == ["one_integers", "two_v_floats", "three_LV", "four_v_LVs"]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy as np
import pytest
import skhep_testdata
from numpy.testing import assert_array_equal

import uproot


def test_numpy():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ.root")) as f:
        a = f["events/Muon_Px"].array(entry_start=1, entry_stop=1, library="np")
        assert isinstance(a, np.ndarray)
        assert len(a) == 0


def test_awkward():
    awkward = pytest.importorskip("awkward")

    with uproot.open(skhep_testdata.data_path("uproot-HZZ.root")) as f:
        a = f["events/Muon_Px"].array(entry_start=1, entry_stop=1)
        assert isinstance(a, awkward.Array)
        assert len(a) == 0



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot


def test_axis():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        assert (
            f["hpx"].axes[0]
            == f["hpx"].axis(0)
            == f["hpx"].axis(-1)
            == f["hpx"].axis("x")
        )
        axis = f["hpx"].axis()
        assert len(axis) == 100
        assert axis[0] == (-4.0, -3.92)
        assert axis[1] == (-3.92, -3.84)
        assert [tuple(x) for x in axis][:3] == [
            (-4.0, -3.92),
            (-3.92, -3.84),
            (-3.84, -3.76),
        ]
        assert axis == axis
        assert f["hpxpy"].axis(0) == f["hpxpy"].axis(1)
        assert axis.traits.circular is False
        assert axis.traits.discrete is False
        assert axis.low == -4
        assert axis.high == 4
        assert axis.width == 0.08
        assert axis.labels() is None
        assert axis.edges()[:5].tolist() == [-4.0, -3.92, -3.84, -3.76, -3.68]
        assert axis.edges(True)[:4].tolist() == [-numpy.inf, -4.0, -3.92, -3.84]
        assert axis.intervals()[:2].tolist() == [[-4.0, -3.92], [-3.92, -3.84]]
        assert axis.intervals(True)[:2].tolist() == [[-numpy.inf, -4.0], [-4.0, -3.92]]
        assert axis.centers()[:4].tolist() == [-3.96, -3.88, -3.8, -3.7199999999999998]
        assert axis.centers()[:3].tolist() == [-3.96, -3.88, -3.8]
        assert axis.centers(True)[:3].tolist() == [-numpy.inf, -3.96, -3.88]
        assert axis.widths()[:3].tolist() == [0.08, 0.08, 0.08]
        assert axis.widths(True)[:2].tolist() == [numpy.inf, 0.08000000000000007]
        assert (
            len(axis)
            == len(axis.edges()) - 1
            == len(axis.intervals())
            == len(axis.centers())
            == len(axis.widths())
        )
        assert (
            len(axis) + 2
            == len(axis.edges(flow=True)) - 1
            == len(axis.intervals(flow=True))
            == len(axis.centers(flow=True))
            == len(axis.widths(flow=True))
        )

    with uproot.open(skhep_testdata.data_path("uproot-issue33.root")) as f:
        assert f["cutflow"].axes[0] == f["cutflow"].axis(0) == f["cutflow"].axis("x")
        axis = f["cutflow"].axis()
        assert len(axis) == 7
        assert axis[0] == "Dijet"
        assert axis[1] == "MET"
        assert list(axis)[:3] == ["Dijet", "MET", "MuonVeto"]
        assert axis == axis
        assert axis.traits.circular is False
        assert axis.traits.discrete is True
        assert axis.low == 0.0
        assert axis.high == 7.0
        assert axis.width == 1.0
        assert axis.labels()[:3] == ["Dijet", "MET", "MuonVeto"]
        assert list(axis) == axis.labels()
        assert axis.edges()[:5].tolist() == [0, 1, 2, 3, 4]
        assert axis.edges(True)[:4].tolist() == [-numpy.inf, 0, 1, 2]
        assert axis.intervals()[:2].tolist() == [[0, 1], [1, 2]]
        assert axis.intervals(True)[:2].tolist() == [[-numpy.inf, 0], [0, 1]]
        assert axis.centers()[:4].tolist() == [0.5, 1.5, 2.5, 3.5]
        assert axis.centers(True)[:3].tolist() == [-numpy.inf, 0.5, 1.5]
        assert axis.widths()[:3].tolist() == [1, 1, 1]
        assert axis.widths(True)[:3].tolist() == [numpy.inf, 1, 1]
        assert (
            len(axis.edges()) - 1
            == len(axis.intervals())
            == len(axis.centers())
            == len(axis.widths())
        )
        assert (
            len(axis.edges(flow=True)) - 1
            == len(axis.intervals(flow=True))
            == len(axis.centers(flow=True))
            == len(axis.widths(flow=True))
        )


def test_bins():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        for i in range(2):
            hpx = f["hpx"]
            hpxpy = f["hpxpy"]
            hprof = f["hprof"]
            assert len(hpx.axis().centers()) == len(hpx.values())
            assert len(hpx.axis().centers(flow=True)) == len(hpx.values(flow=True))
            assert len(hprof.axis().centers()) == len(hprof.values())
            assert len(hprof.axis().centers(flow=True)) == len(hprof.values(flow=True))
            assert (
                len(hpxpy.axis(0).centers()),
                len(hpxpy.axis(1).centers()),
            ) == hpxpy.values().shape
            assert (
                len(hpxpy.axis(0).centers(flow=True)),
                len(hpxpy.axis(1).centers(flow=True)),
            ) == hpxpy.values(flow=True).shape
            assert numpy.all(hpxpy.values() == hpxpy.variances())
            assert numpy.all(hpxpy.values(flow=True) == hpxpy.variances(flow=True))


def test_boost():
    boost_histogram = pytest.importorskip("boost_histogram")

    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        hpx = f["hpx"]
        hpxpy = f["hpxpy"]
        hprof = f["hprof"]

        assert hpx.to_boost() == boost_histogram.Histogram(hpx)

        assert hpx.to_boost().name == "hpx"
        assert hpx.to_boost().axes[0].name == "xaxis"

        assert hpxpy.to_boost().name == "hpxpy"
        assert hpxpy.to_boost().axes[0].name == "xaxis"
        assert hpxpy.to_boost().axes[1].name == "yaxis"

        assert hprof.to_boost().name == "hprof"
        assert hprof.to_boost().axes[0].name == "xaxis"


@pytest.mark.skip(
    reason="Something's wrong with uproot-issue33.root and boost-histogram"
)
def test_boost_2():
    boost_histogram = pytest.importorskip("boost_histogram")

    with uproot.open(skhep_testdata.data_path("uproot-issue33.root")) as f:
        f["cutflow"].to_boost()
        # assert f["cutflow"].to_boost().name == "cutflow"
        # assert f["cutflow"].to_boost().title == "dijethad"
        # assert f["cutflow"].to_boost().axes[0].name == "xaxis"
        # assert f["cutflow"].to_boost().axes[0].title == ""


def test_issue_0722():
    boost_histogram = pytest.importorskip("boost_histogram")

    with uproot.open(skhep_testdata.data_path("uproot-issue-722.root")) as f:
        f["hist"].to_boost()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test():
    with uproot.open(
        skhep_testdata.data_path("uproot-Zmumu-uncompressed.root")
    ) as file:
        basket = file["events/px1"].basket(0)

    # without PR #519, this would be a segfault
    assert basket.data[0] == 192



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import array
import os

import numpy as np
import pytest
import skhep_testdata

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")
awkward = pytest.importorskip("awkward")


def test_awkward_as_numpy(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        fout.mktree("tree", {"b1": "var * float64", "b2": "var * float64"})
        fout["tree"].extend(
            {
                "b1": [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
                "b2": np.array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], dtype="O"),
            }
        )

    with uproot.open(newfile) as fin:
        assert fin["tree/b1"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
        assert fin["tree/b2"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_1(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        fout.mktree("tree", {"nb": "int32", "b": "var * float64"})
        with pytest.raises(ValueError):
            fout["tree"].extend(
                {"nb": [1, 2, 3], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]}
            )
        fout["tree"].extend({"nb": [3, 0, 2], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_2(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        fout.mktree("tree", {"nb": "int32", "b": "var * float64"})
        with pytest.raises(ValueError):
            fout["tree"].extend(
                {"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [1, 2, 3]}
            )
        fout["tree"].extend({"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [3, 0, 2]})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_3(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        fout.mktree("tree", {"b": "var * float64", "nb": "int32"})
        with pytest.raises(ValueError):
            fout["tree"].extend(
                {"nb": [1, 2, 3], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]}
            )
        fout["tree"].extend({"nb": [3, 0, 2], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_4(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        fout.mktree("tree", {"b": "var * float64", "nb": "int32"})
        with pytest.raises(ValueError):
            fout["tree"].extend(
                {"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [1, 2, 3]}
            )
        fout["tree"].extend({"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [3, 0, 2]})

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_5(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        with pytest.raises(ValueError):
            fout["tree"] = {"nb": [1, 2, 3], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]}
        fout["tree"] = {"nb": [3, 0, 2], "b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]]}

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_counter_shadows_branch_6(tmp_path):
    newfile = os.path.join(tmp_path, "newfile.root")

    with uproot.recreate(newfile) as fout:
        with pytest.raises(ValueError):
            fout["tree"] = {"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [1, 2, 3]}
        fout["tree"] = {"b": [[1.1, 2.2, 3.3], [], [4.4, 5.5]], "nb": [3, 0, 2]}

    with uproot.open(newfile) as fin:
        assert fin["tree/nb"].array().tolist() == [3, 0, 2]
        assert fin["tree/b"].array().tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
